package CAFI;

/**
AgentInSpace.java    extends Agent

A class that can be root class for agents that can move in GridWorld, an 
extension of the Repast Object2DGrid class.
Note it implements the Repast Drawable interface, so that
it can be displayed via the Repast Object2DDisplay gui object,
and the ObjectInGrid interface, so it can be used in GridWorld's.

**/

import java.util.Formatter;
import java.awt.Point;
import java.util.ArrayList;

import uchicago.src.sim.util.*;
import uchicago.src.sim.gui.*;
import uchicago.src.sim.space.*;
import uchicago.src.sim.gui.ColorMap;
import java.awt.Color;

public class AgentInSpace extends Agent implements Drawable, ObjectInGrid {
	// class field
    public  static int          	nextId = 0; // to give each a unique id

	// fields (instance variables)  
	public int 			x, y;	// cache the x,y location
	public Color		myColor;    

	// an constructor
	public AgentInSpace ( ) {
		id = nextId++;
		x = 0;		y = 0;
		live = true;
		myColor = Color.orange; 	// default color;
	}


	////////////////////////////////////////////////////////////////////////////
	// step
	// Top level definition of what the agent can do each time its activiated.
	// Currently it does this:
	// - get random dx,dy values in -1,0,1 as possible move to make
	// - asks the world to move it by the selected amounts (dx,dy)
	//   world may not move it for various reasons...

	public void step () {
		boolean moved = false;  // not moved this step so far

		if ( model.getRDebug() > 0 ) 
			System.err.printf( "   --AgentInSpace-step() for id=%d at x,y=%d,%d.\n",
						   id, x, y );

		// get a random amout to move into dx,dy
		int dx = Model.getUniformIntFromTo( -1, 1 );  // dx = { -1,0,1 } 
		int dy = Model.getUniformIntFromTo( -1, 1 );  // dy = { -1,0,1 }

		if ( model.getRDebug() > 0 ) 
			System.err.printf( "   - try to move dx,dy = %d,%d\n", dx, dy );
		Object2DGridXMove move = world.moveObject( this, dx, dy );
		if ( move == Object2DGridXMove.MOVEOFFWORLD ) {
		    if ( model.getRDebug() > 0 )
				System.out.printf( "   *** tried to move offworld.\n" );
		}
		else if ( move == Object2DGridXMove.MOVETOOCCUPIEDCELL ) {
			if ( model.getRDebug() > 0 )
				System.out.printf( "   *** tried to move to occupied cell.\n" );
		}
		else { // moved ok!
			moved = true;
		}

		if ( model.getRDebug() > 1 ) 
			System.err.printf("      AgentInSpace.step() done. moved=%b to %d,%d.\n", 
							  moved, x, y );

	}


	////////////////////////////////////////////////////////////
	// printOut
	// print fields to System.out.
	//
	// Note: we can use the class's field variables in all methods.

	public void printOut ( ) {
		System.out.printf( " - AgentInSpace id %2d (x,y=%d,%d; live=%b).\n",
						   id, x, y, live );
	}

	////////////////////////////////////////////////////////////////////////////
	// setters and getters
	//
	public int getX() { return x; }
	public void setX( int i ) { x = i; }
	public int getY() { return y; }
	public void setY( int i ) { y = i; }

	public void setMyColor ( Color c ) { myColor = c; }

	// note these are class (static) methods, to set class (static) variables
	public static void setNextId ( int i ) { nextId = i; }

	////////////////////////////////////////////////////////////////////////////
	// draw 
	// we implement Drawable interface, so we need this method
	// so that the ant can draw itself when requested  (by the GUI display).

    public void draw( SimGraphics g ) {
		g.drawFastRoundRect( myColor );
    }


}
