package CAFI;


import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.Color;
import java.util.ArrayList;
import java.util.Formatter;
import java.math.*;

import uchicago.src.sim.engine.Schedule;
import uchicago.src.sim.engine.SimModelImpl;
import uchicago.src.sim.engine.*;
import uchicago.src.sim.util.*;
import uchicago.src.sim.gui.*;
import uchicago.src.sim.gui.ColorMap;
import uchicago.src.sim.gui.DisplaySurface;
import uchicago.src.sim.gui.Object2DDisplay;
import uchicago.src.sim.space.Object2DGrid;
import uchicago.src.sim.gui.Value2DDisplay;
import uchicago.src.sim.space.Object2DTorus;
import uchicago.src.sim.engine.AbstractGUIController;
import uchicago.src.sim.analysis.*;


public class GUIModel extends Model {

    private Object2DDisplay  worldDisplay;	// 2D Object lattice -> display (Repast)
    private ArrayList<Value2DDisplay> speciesVolume;
    private Value2DDisplay roads;
    private Value2DDisplay roadDistance;
    private Value2DDisplay roadTransportDistance;
    private Value2DDisplay loggerKnownNetValue;
    private Value2DDisplay monitorKnownSpecies2;
    private Value2DDisplay newRoadValues;
    private Value2DDisplay cutSp1OrBoth;
    private DisplaySurface	 dsurf;		    // display surface (RePast)	
    private DisplaySurface	 dsurf2;		    // display surface (RePast)	
    private DisplaySurface	 dsurf3;		    // display surface (RePast)
    private DisplaySurface	 dsurf4;		    // display surface (RePast)	
    private DisplaySurface   dsurf5;
    
    public  OpenSequenceGraph		graph;
    public  OpenSequenceGraph		graph2;
    public  OpenSequenceGraph		graph3;
    public  OpenSequenceGraph		graph4;  
    public  OpenSequenceGraph		graph5;  

	public static ColorMap		 colorMapTree;
	public static ColorMap		 colorMapRoad;
	public static ColorMap		 colorMapCost;
	public static ColorMap		 colorMapTransport;
	
	public static int 			 colorMapTreeSize = 10; //dummy value - this gets reset later
	public static double 		 colorMapTreeMax = 9;  //dummy value - this gets reset later
	public static final int      colorMapRoadSize = 10;
	public static final double   colorMapRoadMax =  colorMapRoadSize - 1.0;
	public static final int      colorMapTransportSize = 200;
	public static final double   colorMapTransportMax =  colorMapRoadSize - 1.0;
	public static final int      colorMapCostSize = 501;
	public static final double   colorMapCostMax =  (int) colorMapCostSize/2;
	
	/////////////////////////////////////////////////////////////////////
	// setup
	//
	// this runs automatically when the model starts
	// and when you click the reload button, to "tear down" any 
	// existing display objects, and get ready to initialize 
	// them at the start of the next 'run'.
	//
	public void setup() {
	   	//System.out.printf( "==> GUIModel setup() called...\n" );

		super.setup();  // the super class does conceptual-model setup

		// NOTE: you may want to set these next two to 'true'
		// if you are on a windows machine.  that would tell repast
		// to by default send System.out and .err output to
		// a special repast output window.
		AbstractGUIController.CONSOLE_ERR = false;
		AbstractGUIController.CONSOLE_OUT = false;

		if ( dsurf != null ) dsurf.dispose();
		if ( dsurf2 != null ) dsurf2.dispose();
		if ( dsurf3 != null ) dsurf3.dispose();
		if ( dsurf4 != null ) dsurf4.dispose();
		if ( dsurf5 != null ) dsurf5.dispose();

		if (graph != null)  graph.dispose();
		if (graph2 != null)  graph2.dispose();
		if (graph3 != null)  graph3.dispose();
		if (graph4 != null)  graph4.dispose();
		if (graph5 != null)  graph5.dispose();
		graph = null;
		graph2 = null;
		graph3 = null;
		graph4 = null;
		graph5 = null;

		// create the ColorMap for showing tree volume in green
		// Note this probably should be in buildDisplay(), to allow
		// for parameterized control of the map creation.
		
		//set the colormap for tree here, because it is dependent on what the biomass of the forest is - if the volume
		//of forest in a cell is really low, have to have a small colormap (silly value2ddisplay only takes int values
		//and it doesn't seem smart to set up a whole other display with values multiplied through by a scalar just for 
		//display
		int highestVolume = (int) Math.max(maxVolumeS1, Math.max(maxVolumeS2, maxVolumeS0));
		int      colorMapTreeSize = Math.min(64,highestVolume) ;
		double   colorMapTreeMax =  colorMapTreeSize - 1.0;
		
		colorMapTree = new ColorMap ();
		for (int i = 0; i < colorMapTreeSize; i++) {
            // we specify the position i, and a fraction of each of RGB shade
			colorMapTree.mapColor ( i, 0, i /  colorMapTreeMax, 0);
		}

		// create the ColorMap for showing roads in brown
		// Note this probably should be in buildDisplay(), to allow
		// for parameterized control of the map creation.
		colorMapRoad = new ColorMap ();
		for (int i = 0; i < colorMapRoadSize; i++) {
            // we specify the position i, and a fraction of each of RGB shade
			colorMapRoad.mapColor ( i, 3*i /  colorMapRoadMax, 2*i / colorMapRoadMax, i /colorMapRoadMax);
		}

		// create the ColorMap for showing transport costs in brown
		// Note this probably should be in buildDisplay(), to allow
		// for parameterized control of the map creation.
		colorMapTransport = new ColorMap ();
		for (int i = 0; i < colorMapTransportSize; i++) {
            // we specify the position i, and a fraction of each of RGB shade
			colorMapTransport.mapColor ( i, 3*i /  colorMapTransportMax, 2*i / colorMapTransportMax, i /colorMapTransportMax);
		}
		
		// create the ColorMap for showing costs in red and green
		// Note this probably should be in buildDisplay(), to allow
		// for parameterized control of the map creation.
		
		// have to do a bit of a hack to make it work!!!!! ie, everything is shifted UP by maxCost to make things all positive
		colorMapCost = new ColorMap ();
		for (int i = - (int) colorMapCostSize/2; i < 0; i++) {
            // we specify the position i, and a fraction of each of RGB shade
			colorMapCost.mapColor ( i, - i /  colorMapCostMax, 0, 0);
		}
		for (int i = 0; i < (int) colorMapCostSize/2; i++) {
            // we specify the position i, and a fraction of each of RGB shade
			colorMapCost.mapColor ( i, 0, i /  colorMapCostMax, 0);
		}		
		dsurf = null;
		dsurf2 = null;
		dsurf3 = null;
		dsurf4 = null;
		dsurf5 = null;

		if ( rDebug > 0 )
			System.out.printf( "<== GUIModel setup() done.\n" );
	}

	/////////////////////////////////////////////////////////////////////
	// begin
	//
	// this runs when you click the "initialize" button
	// (the button with the single arrow that goes around in a circle)
	//
	public void begin()	{
		DMSG(1, "==> enter GUIModel-begin()" );
		buildModel();     			// the base model does this
		buildDisplay();
		buildSchedule();
		DMSG(1, "<== leave GUIModel-begin() done." );
	}

	/////////////////////////////////////////////////////////////////////
	// buildDisplay
	//
	// builds the display and display related things
	//
	public void buildDisplay() {
		if ( rDebug > 0 )
			System.out.printf( "==> GUIModel buildDisplay...\n" );

		worldDisplay = new Object2DDisplay( world );
		
		dsurf2 = new DisplaySurface( this, "Roads" );
		registerDisplaySurface( "Road Display", dsurf2 );
        addSimEventListener( dsurf2 );  // link to the other parts of the repast gui
		
		dsurf3 = new DisplaySurface( this, "Trees" );
		registerDisplaySurface( "Tree Display", dsurf3 );
        addSimEventListener( dsurf3 );  // link to the other parts of the repast gui

		dsurf4 = new DisplaySurface( this, "Knowledge" );
		registerDisplaySurface( "Knowledge Display", dsurf4 );
        addSimEventListener( dsurf4 );  // link to the other parts of the repast gui
        
        dsurf5 = new DisplaySurface( this, "Cut Sp1 or Both" );
		registerDisplaySurface( "Sp1 Sp2 Display", dsurf5 );
        addSimEventListener( dsurf5 );  // link to the other parts of the repast gui
		
		roads = new Value2DDisplay( roadNetwork.roads, colorMapRoad );
		roadDistance = new Value2DDisplay( roadNetwork.distanceToRoads, colorMapRoad );
		roadTransportDistance = new Value2DDisplay( roadNetwork.cumMaintenanceCostToAccess, colorMapTransport);


		
		int mTree = (int) (Math.max(maxVolumeS1, Math.max(maxVolumeS2, maxVolumeS0)) / colorMapTreeMax);
		speciesVolume = new ArrayList<Value2DDisplay> ();
		for(int i = 0; i < numSpecies; ++i){
			
			Value2DDisplay newSpeciesVolume = new Value2DDisplay( ecosystem.getTree(i), colorMapTree );
		
			newSpeciesVolume.setDisplayMapping( mTree, 0 );
			
			speciesVolume.add(newSpeciesVolume);
			String data = "Tree Species " + i;
			dsurf3.addDisplayable( newSpeciesVolume, data);	
		}
		
		int mCost = (int) (maxCost / colorMapCostMax);

		roadTransportDistance.setDisplayMapping(mCost, 0);
		
		loggerKnownNetValue = new Value2DDisplay( loggerList.get(0).canSeeGrid, colorMapCost);
		loggerKnownNetValue.setDisplayMapping(mCost, 0);
		dsurf4.addDisplayableProbeable(loggerKnownNetValue, "Logger Known Net Value");

		monitorKnownSpecies2 = new Value2DDisplay( localEnforcersList.get(0).canSeeGrid, colorMapTree);
		monitorKnownSpecies2.setDisplayMapping(mTree, 0);
		//dsurf4.addDisplayable(monitorKnownSpecies2, "Monitor Known Species 2");
		
		newRoadValues = new Value2DDisplay(loggerList.get(0).newRoadValues, colorMapCost);
		newRoadValues.setDisplayMapping(mCost, 0);
		dsurf4.addDisplayableProbeable(newRoadValues, "Potential Value of New Road Nodes");
		
		cutSp1OrBoth = new Value2DDisplay(loggerList.get(0).cutSp1OrBoth, colorMapCost);
		cutSp1OrBoth.setDisplayMapping(mCost, 0);
		dsurf5.addDisplayableProbeable(cutSp1OrBoth, "Cut Just Sp1 or Both?");
		
		dsurf2.addDisplayableProbeable( roadDistance, "Road Distances");
		dsurf2.addDisplayable( roads, "Road Network");
		//dsurf2.addDisplayableProbeable( roadTransportDistance, "Road Transport Length");
		
		dsurf2.display();
		dsurf3.display();
		dsurf4.display();
		dsurf5.display();
		
		// set up sequence for ant pop size, etc, and graph to plot on
		class SeqLocalEnforcersSize implements Sequence {
			public double getSValue() {
				return localEnforcersList.size();
			}
		}
		class SeqLocalEnforcersAvgX implements Sequence {
			public double getSValue() {
				return getLocalEnforcersAvgX();
			}
		}
		class LoggerNetWorth implements Sequence {
			public double getSValue() {
				return loggerList.get(0).netWorth;
			}
		}
		class MonitorFinesCollected implements Sequence {
			public double getSValue() {
				return localEnforcersList.get(0).finesCollected;
			}
		}
		class LoggerBribeFines implements Sequence {
			public double getSValue() {
				return loggerList.get(0).bribeFinesPaid;
			}
		}
		class MonitorBribes implements Sequence {
			public double getSValue() {
				return localEnforcersList.get(0).bribesCollected;
			}
		}
		class Species1Frac implements Sequence {
			public double getSValue() {
				return (double) currentBiomass[1] / initialBiomass[1];
			}
		}
		class Species2Frac implements Sequence {
			public double getSValue() {
				return (double) currentBiomass[2] / initialBiomass[2];
			}
		}
		
		class LoggerLoggingFine implements Sequence {
			public double getSValue() {
				return (double) loggerList.get(0).percRisk;
			}
		}
		class LoggerBribingFine implements Sequence {
			public double getSValue() {
				return (double) loggerList.get(0).m;
			}
		}
		class MonitorBribingFine implements Sequence {
			public double getSValue() {
				return (double) localEnforcersList.get(0).m;
			}
		}
		
		class RoadBranchPoints implements Sequence {
			public double getSValue() {
				return (double) roadNetwork.branchPoints;
			}
		}
		
		class MinMaxRatio implements Sequence {
			public double getSValue() {
				return (double) localEnforcersList.get(0).currentMaxBribes/localEnforcersList.get(0).currentMinBribes;
			}
		}
		
		graph = new OpenSequenceGraph( "Money earned, Money taken away", this );
		graph.setXRange( 0, 200 );
		graph.setYRange( 0, 100 );
		graph.setYIncrement( 20 );
		graph.setAxisTitles( "Time", "Net Worth" );
		graph.addSequence("Logger Net Worth", new LoggerNetWorth (), Color.BLACK );
		graph.addSequence("Harvesting Fines Collected", new MonitorFinesCollected (), Color.RED );
		graph.addSequence("Bribes Paid" , new MonitorBribes(), Color.GREEN);
		graph.addSequence("Bribing Fines Collected" , new LoggerBribeFines(), Color.BLUE);
		
		graph.display();

		graph2 = new OpenSequenceGraph( "Forest Biomass Fractions Standing", this );
		graph2.setXRange( 0, 200 );
		graph2.setYRange( 0, 1 );
		graph2.setYIncrement( .1 );
		graph2.setAxisTitles( "Time", "Fraction" );
		graph2.addSequence("Species 1", new Species1Frac (), Color.BLACK );
		graph2.addSequence("Species 2", new Species2Frac (), Color.RED );
		
		graph2.display();
		
		graph3 = new OpenSequenceGraph( "Perceived Probabilities", this );
		graph3.setXRange( 0, 200 );
		graph3.setYRange( 0, 1 );
		graph3.setYIncrement( 0.1 );
		graph3.setAxisTitles( "Time", "Probability" );
		graph3.addSequence("Logger Risk of Logging Fines", new LoggerLoggingFine (), Color.BLACK );
		graph3.addSequence("Logger Risk of Bribing Fines", new LoggerBribingFine(), Color.RED );
		graph3.addSequence("Monitor Risk of Bribing Fines", new MonitorBribingFine(), Color.GREEN );
		
		graph3.display();
		
		graph4 = new OpenSequenceGraph( "Bribe Window", this );
		graph4.setXRange( 0, 200 );
		graph4.setYRange( 0, 50 );
		graph4.setYIncrement( 1 );
		graph4.setAxisTitles( "Time", "Ratio" );
		//graph4.setXViewPolicy(OpenSequenceGraph.SHOW_LAST);
		graph4.setYViewPolicy(OpenSequenceGraph.SHOW_FIRST);
		graph4.addSequence("MaxBribe/MinBribe", new MinMaxRatio (), Color.BLACK );

		
		graph4.display();
		
		graph5 = new OpenSequenceGraph( "Branch Points", this );
		graph5.setXRange( 0, 200 );
		graph5.setYRange( 0, 100 );
		graph5.setYIncrement( 1 );
		graph5.setAxisTitles( "Time", "Number of Points" );
		//graph5.setXViewPolicy(OpenSequenceGraph.SHOW_LAST);
		graph5.setYViewPolicy(OpenSequenceGraph.SHOW_FIRST);
		graph5.addSequence("Road Branch Points", new RoadBranchPoints (), Color.BLACK );

		
		graph5.display();
		
		if ( rDebug > 0 )
			System.out.printf( "<== GUIModel buildDisplay done.\n" );
	}

	////////////////////////////////////////////////////////////////
	// buildSchedule
	//
	// This builds the entire schedule, i.e., 
	//  - the base model step (calls stepReport)
	//  - display steps.

	public void buildSchedule() {

		if ( rDebug > 0 )
			System.out.printf( "==> GUIModel buildSchedule...\n" );

		// schedule the current GUIModel's step() function
		// to execute every time step starting with time step 0
		schedule.scheduleActionBeginning( 0, this, "step" );

		// schedule the current GUIModel's processEndOfRun() 
		// function to execute at the end of the run
		schedule.scheduleActionAtEnd( this, "processEndOfRun" );
	}


	///////////////////////////////////////////////////////////////////////////////
	// step
	//
	// executed each step of the model.
	// Ask the super-class to do its step() method,
	// and then this does display related activities.
	//
	public void step() {

		super.step();			// the model does whatever it does

		numSteps++;
		
		
		if( getTickCount() % cuttingDays == 0 && getTickCount() > 1 ) {
						
			
			roads = new Value2DDisplay( roadNetwork.roads, colorMapRoad );
			roadDistance = new Value2DDisplay( roadNetwork.distanceToRoads, colorMapRoad );


			
			int mTree = (int) (Math.max(maxVolumeS1, Math.max(maxVolumeS2, maxVolumeS0)) / colorMapTreeMax);
			speciesVolume = new ArrayList<Value2DDisplay> ();
			for(int i = 0; i < numSpecies; ++i){
				
				Value2DDisplay newSpeciesVolume = new Value2DDisplay( ecosystem.getTree(i), colorMapTree );
			
				newSpeciesVolume.setDisplayMapping( mTree, 0 );
				
				speciesVolume.add(newSpeciesVolume);
				String data = "Tree Species " + i + ",Year " + Math.floor(numSteps/cuttingDays);
				dsurf3.addDisplayable( newSpeciesVolume, data);	
			}
			

			int mCost = (int) (maxCost / colorMapCostMax);

			
			loggerKnownNetValue = new Value2DDisplay( loggerList.get(0).canSeeGrid, colorMapCost);
			loggerKnownNetValue.setDisplayMapping(mCost, 0);
			dsurf4.addDisplayable(loggerKnownNetValue, "Logger Known Net Value, Year " + Math.floor(numSteps/cuttingDays));

			monitorKnownSpecies2 = new Value2DDisplay( localEnforcersList.get(0).canSeeGrid, colorMapTree);
			monitorKnownSpecies2.setDisplayMapping(mTree, 0);
			dsurf4.addDisplayable(monitorKnownSpecies2, "Monitor Known Species 2, Year " + Math.floor(numSteps/cuttingDays));
			
			
			dsurf2.addDisplayable( roadDistance, "Road Distances, Year " + Math.floor(numSteps/cuttingDays));
			dsurf2.addDisplayable( roads, "Road Network, Year " + Math.floor(numSteps/cuttingDays));
		
		}
		
		if(Math.IEEEremainder(numSteps, graphUpdateFrequency) == 0) {
			// add things after this for all displays (graphs, etc)
			//dsurf.updateDisplay();
			dsurf2.updateDisplay();
			dsurf3.updateDisplay();
			dsurf4.updateDisplay();
			dsurf5.updateDisplay();
			graph.step();
			graph2.step();
			graph3.step();
			graph4.step();
			graph5.step();
			}

	}

	// processEndOfRun
	// called once, at end of run.
	public void processEndOfRun ( ) {
		if ( rDebug > 0 )  
			System.out.printf("\n\n===== GUIModel processEndOfRun =====\n\n" );
		applyAnyStoredChanges();
		endReportFile();
		this.fireStopSim();
	}


/////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////
//   ****  NO NEEd TO CHANGE THE REST OF THIS  *****

	////////////////////////////////////////////////////////////////////
	// main entry point
	public static void main( String[] args ) {

		uchicago.src.sim.engine.SimInit init =
			new uchicago.src.sim.engine.SimInit();
		GUIModel model = new GUIModel();

		//System.out.printf("==> GUIMOdel main...\n" );

		// set the type of model class, this is necessary
		// so the parameters object knows whether or not
		// to do GUI related updates of panels,etc when a
		// parameter is changed
		model.setModelType("GUIModel");

        // Do this to set the Update Probes option to true in the
        // Repast Actions panel
        Controller.UPDATE_PROBES = true;

		model.setCommandLineArgs( args );
		init.loadModel( model, null, false ); // does setup()

		// this new function calls ProbeUtilities.updateProbePanels() and 
		// ProbeUtilities.updateModelProbePanel()
		model.updateAllProbePanels();

	}

}
