package CAFI;

/**
LocalEnforcer.java 

Class for local enforcers, a type of AgentInSpace

**/

import java.util.Collections;
import java.util.Comparator;
import java.util.Formatter;
import java.util.Vector;
import java.awt.Point;
import java.util.ArrayList;
import java.lang.Math;

import uchicago.src.sim.util.*;
import uchicago.src.sim.gui.*;
import uchicago.src.sim.space.*;
import uchicago.src.sim.gui.ColorMap;
import java.awt.Color;


public class LocalEnforcer extends Agent {
	// "class" variables -- one value for all instances 
    public  static int          	nextId = 0; // to give each a unique id

	// fields (instance variables)  
	int 				id;		// unique id number for each instance


	public Vector<Observation> monitoringSites;
	public Vector<Report> reportSites;
	public Object2DGrid canSeeGrid;
	public Object2DGrid observedCells;
	public double		bribesCollected;
	public double		bribeFinesPaid;
	public double		finesCollected;
	public int   sumNumTimesBribed;
	public int	sumNumTimesBribeCaught;
	public int[] 		numTimesBribed;
	public int[]		numTimesBribeCaught;
	public double 		m;
	public double		observedIllegalCutting;
	public double		currentMaxBribes;
	public double		currentMinBribes;
	public int			currentNumBribes;

	// a constructor
	public LocalEnforcer ( ) {
		id = nextId++;
		
		finesCollected = 0;
		canSeeGrid = new Object2DGrid (model.sizeX, model.sizeY);
		monitoringSites = new Vector<Observation> ();
		reportSites = new Vector<Report> ();
		canSeeGrid = new Object2DGrid(model.sizeX, model.sizeY);
		observedCells = new Object2DGrid(model.sizeX, model.sizeY);
		for(int i = 0; i < model.sizeX; i++) {
			for(int j = 0; j < model.sizeY; j++) {
				canSeeGrid.putValueAt(i, j, 0);	
				observedCells.putValueAt(i, j, 0);
			}
		}
		

		numTimesBribed = new int[model.riskMemory] ;
		numTimesBribeCaught = new int[model.riskMemory];
		for(int i = 0; i < model.riskMemory; ++i) {
			numTimesBribed[i] = 0;
			numTimesBribeCaught[i] = 0;
		}
		numTimesBribed[model.riskMemory-1] = 1;  //so that getting caught on the first bribe won't keep them from doing it
		
		m = 0;
		bribesCollected = 0;
		bribeFinesPaid = 0;
		observedIllegalCutting = 0;
	}

	////////////////////////////////////////////////////////////////////////////
	// step
	// Top level definition of what the agent can do each time its activated.

	public void step () {
		if ( model.getRDebug() > 0 ) 
			System.err.printf( "   -->NGO-step() for id=%d.\n", id );

		resetAverages();
		recalculateRisk();
		
		lookForest ();

		trimMonitoringSitesMemory ();

		if ( model.getRDebug() > 1 ) 
			System.err.printf("    <--NGO-step() done.\n" );
	}

	////////////////////////////////////////////////////////////
	//revisit our risk perceptions
	public void resetAverages() {
		currentMaxBribes = 0;
		currentMinBribes = 0;
		currentNumBribes = 0;
	}
	////////////////////////////////////////////////////////////
	//revisit our risk perceptions
	public void recalculateRisk() {

		
		sumNumTimesBribed = numTimesBribed[0];
		sumNumTimesBribeCaught = numTimesBribeCaught[0];
		for(int i = 1; i < model.riskMemory; ++i) {
			sumNumTimesBribed += numTimesBribed[i];
			sumNumTimesBribeCaught += numTimesBribeCaught[i];
			numTimesBribed[i-1] = numTimesBribed[i];
			numTimesBribeCaught[i-1] = numTimesBribeCaught[i];
		}
		numTimesBribed[model.riskMemory-1] = 0;
		numTimesBribeCaught[model.riskMemory-1] = 0;
		
		if(sumNumTimesBribed > 0) {
			m = (double) sumNumTimesBribeCaught/sumNumTimesBribed;
		} else {
			m = 0;
			numTimesBribed[model.riskMemory-1] = 1;  //so that getting caught on the first bribe won't keep them from doing it
		}
	}
	
	////////////////////////////////////////////////////////////
	//look at the forest to see where we should monitor
	public void lookForest() {


		//for now we are working with just one logger - needs to be updated if this changes
		Logger currentLogger = model.loggerList.get(0);
		//make sure we don't try to monitor more cells than we know about
		int limit = 0;
		if (model.maxMonCells > monitoringSites.size()) {
			limit = monitoringSites.size();
		}else {
			limit = model.maxMonCells;
		}
		
		//checkpoint monitoring
		double currentLoggedSp1 = model.loggerList.get(0).currentLoggedSp1;
		double currentLoggedSp2 = model.loggerList.get(0).currentLoggedSp2;
		
		double r = model.loggerList.get(0).currentLoggedSp1/(model.loggerList.get(0).currentLoggedSp1 + model.loggerList.get(0).currentLoggedSp2);
		if(r*model.maxProbCheckpointCatch > Model.getUniformDoubleFromTo(0, 1)) {
			//checkpoint caught the illegal logging
			currentLogger.netWorth = currentLogger.netWorth - currentLoggedSp2 * model.fineSpecies2PerM3;
			currentLogger.harvestFinesPaid += currentLoggedSp2 * model.fineSpecies2PerM3;
			finesCollected += currentLoggedSp2 * model.fineSpecies2PerM3;
			observedIllegalCutting += currentLoggedSp2;
			
		}
		//site monitoring
		for(int i = 0; i < limit; ++i) {
			
			double oldObsSpecies2 = (double) monitoringSites.get(i).massSpecies2;
			Observation currentObs = monitoringSites.get(i);
			currentObs.updateObservation();
			canSeeGrid.putValueAt(currentObs.x, currentObs.y, currentObs.massSpecies2);
			double currentObsSpecies2 = (double) monitoringSites.get(i).massSpecies2;
			
			//detection, bribing, enforcement

			if (currentObsSpecies2 < oldObsSpecies2 && (model.roadNetwork.roads.getValueAt(currentObs.x, currentObs.y) == 0 || (model.roadNetwork.roads.getValueAt(currentObs.x, currentObs.y) > 0 && !model.allowedCutRoads))) { //trees have been illegally cut!!
				//enter here if there is observed species 2 cutting AND we are either 1) not along a road route or 2) along a road route but not allowed to cut on roads
				
				if(model.allowedCutRoads && model.roadNetwork.roads.getValueAt(currentObs.x, currentObs.y) > 0) {
					int a = 1;
				}
				//calculate the potential revenue for this instance
				double R = (oldObsSpecies2 - currentObsSpecies2)*model.priceSpecies2PerM3;
				
				//calculate m for the logger
				double mLogger = currentLogger.m;
				
				//calc the min bribe the monitor will accept, and the max bribe a logger will pay
				double minBribe = (double) m * model.phiMonitor * R;
				double maxBribe = (double) model.fineSpecies2PerM3 * (oldObsSpecies2 - currentObsSpecies2)*(1-mLogger) - mLogger * model.phiLogger * R;
				currentMaxBribes += maxBribe;
				currentMinBribes += minBribe;
				currentNumBribes++;
				
				double pMonitorAccept = 0;
				double pLoggerAccept = 0;
				double bribe = 0;
				if (minBribe < maxBribe) { //have an envelope of possible bribes
					//estimate the actual bribe negotiated, based on bargaining power and an assumption of normal distribution of bribes
					double baseBribe = minBribe + model.betaBribe*(maxBribe - minBribe);
					bribe = model.getNormalDouble(baseBribe, (double) baseBribe*model.sigmaBribe);
					
					//figure out if it will be accepted, assuming that monitors are more likely to accept
					//as bribe approaches maxBribe, loggers more likely as bribe approaches minBribe
					//**at present, assuming uniform probability change across the envelope
					//**the occurrence of bribes is thus an interaction among the normal distribution of negotiated bribes
					//**and the uniform distribution of this likelihood of accepting -- need to think about this and either
					//**modify, or justify
					pMonitorAccept = (double) Math.max(Math.min((bribe-minBribe)/(baseBribe - minBribe),1.0),0);
					pLoggerAccept = (double) Math.max(Math.min((maxBribe - bribe)/(maxBribe - baseBribe),1.0),0);
				} 
				
				//if both sides accept the bribe, make it so
				if(model.bribeOn == true && Model.getUniformDoubleFromTo(0, 1) < pMonitorAccept && Model.getUniformDoubleFromTo(0, 1) < pLoggerAccept) {
					currentLogger.netWorth = currentLogger.netWorth - bribe; 
					currentLogger.bribesPaid += bribe; 
					currentLogger.numTimesBribed[model.riskMemory-1]++;
					numTimesBribed[model.riskMemory-1]++;
					bribesCollected += bribe;
					Report newReport = new Report(currentObs.x, currentObs.y, 0);
					reportSites.add(newReport);
				} else { //fine the logger as normal
					currentLogger.netWorth = currentLogger.netWorth - (oldObsSpecies2 - currentObsSpecies2) * model.fineSpecies2PerM3;
					currentLogger.harvestFinesPaid += (oldObsSpecies2 - currentObsSpecies2) * model.fineSpecies2PerM3;
					finesCollected += (oldObsSpecies2 - currentObsSpecies2) * model.fineSpecies2PerM3;
					currentLogger.numTimesSpecies2Fined[model.riskMemory-1]++;
					observedIllegalCutting += (oldObsSpecies2 - currentObsSpecies2);
					Report newReport = new Report(currentObs.x, currentObs.y, oldObsSpecies2 - currentObsSpecies2);
					reportSites.add(newReport);
				}
			}

			
		}
		
		//number of new cells to survey
		int numSeeCells = model.monRandCells;
		
		
		//some different options for how monitors pick cells to look at
		
		//1 - probability proportional to distance from road
		Vector<Double> maxDistanceVector = model.roadNetwork.distanceToRoads.findMaximum(model.sizeX/2, model.sizeY/2, Math.max(model.sizeX,model.sizeY), true, Discrete2DSpace.VON_NEUMANN);
		double maxDistance = (double) maxDistanceVector.get(0);
		Object2DGrid probSelection = new Object2DGrid(model.sizeX,model.sizeY);
		double sumDistance = 0;
		for(int i = 0; i < model.sizeX; i++) {
			for(int j = 0; j < model.sizeY; j++) {
				double currentDistance =  model.roadNetwork.distanceToRoads.getValueAt(i, j);
				sumDistance += maxDistance - currentDistance + 1;
				probSelection.putValueAt(i, j, maxDistance - currentDistance + 1);
			}
		}
		for(int i = 0; i < model.sizeX; i++) {
			for(int j = 0; j < model.sizeY; j++) {
				probSelection.putValueAt(i, j, (double) probSelection.getValueAt(i, j) / sumDistance);
			}
		}		
		
		//randomly select a subset of cells to look at
		for(int i = 0; i < numSeeCells; ++i) {

			int newX = ModelParameters.getUniformIntFromTo(0, model.sizeX - 1);
			int newY = ModelParameters.getUniformIntFromTo(0, model.sizeY - 1);
			if(observedCells.getValueAt(newX, newY) == 0) {
				//if they are cells we haven't seen yet, add them to the list
				Observation newObs = new Observation(newX,newY);
				monitoringSites.add(newObs);
				observedCells.putValueAt(newX, newY, 1);
				canSeeGrid.putValueAt(newX, newY, newObs.massSpecies2);
			}
		}
		
		//order them from highest to lowest mass of species 2
		removeDuplicates();
	}

	//sort, and remove duplicates from the vector of memory
	public void removeDuplicates () {
		int i = 0;
		Collections.sort( monitoringSites, (java.util.Comparator<? super Observation>) new ObsSpecies2Comparator() );
		while (i < monitoringSites.size()-1) {
			if(monitoringSites.get(i).x == monitoringSites.get(i+1).x && monitoringSites.get(i).y == monitoringSites.get(i+1).y) {
				monitoringSites.removeElementAt(i+1);
			} else {
				i = i + 1;
			}
		}
	}

	
	//trim down the vector of memory
	public void trimMonitoringSitesMemory () {
		while (monitoringSites.size() > model.lengthMemory) {
			int currentSize = monitoringSites.size()-1;
			int currentX = monitoringSites.get(currentSize).x;
			int currentY = monitoringSites.get(currentSize).y;
			observedCells.putValueAt(currentX, currentY, 0);
			canSeeGrid.putValueAt(currentX, currentY, 0);
			monitoringSites.removeElementAt(currentSize);
			
		}
	}
	
	private class ObsSpecies2Comparator implements Comparator {
	       public final int compare ( Object obs1, Object obs2 ) {
	           double d1 = ( (Observation) obs1 ).massSpecies2;
	           double d2 = ( (Observation) obs2 ).massSpecies2;
	           if ( d1 < d2 ) //orders values descending...
	               return 1;
	           else if ( d1 > d2 )
	               return -1;
	           return 0;
	       }
	   }
	
	////////////////////////////////////////////////////////////
	// printOut
	// print fields to System.out.
	//
	// Note: we can use the class's field variables in all methods.

	public void printOut ( ) {
		System.out.printf( " - LocalEnforcer id %2d (x,y=%d,%d; live=%b).\n",
						   id );
	}


}
