package CAFI;

/**
Logger.java     

A class for NGO agents. 
It extends the generic Agent class, so its not located in the world.


 **/

import java.util.Formatter;
import java.util.Comparator;
import java.awt.Point;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Vector;

import com.bbn.openmap.tools.roads.RoadServices;



import uchicago.src.sim.util.*;
import uchicago.src.sim.gui.*;
import uchicago.src.sim.space.Discrete2DSpace;
import uchicago.src.sim.space.Object2DGrid;

public class Logger extends Agent {
	// "class" variables -- one value for all instances 
	public  static int          	nextId = 0; // to give each a unique id

	// fields (instance variables)  
	int 				id;		// unique id number for each instance

	Vector<Integer>	loggingSitesX;
	Vector<Integer>			loggingSitesY;
	Vector<Double>			loggingSitesNetValue;
	Vector<Observation>		loggingSites;
	Vector<NewRoadObs>		newRoadSites;
	public Object2DGrid	canSeeGrid;
	public Object2DGrid	observedCells;	
	public Object2DGrid  newRoadValues;
	public Object2DGrid  cutSp1OrBoth;
	public double			netWorth;
	public double			loggingRevenue;
	public double			loggingRevenueSp1;
	public double			loggingCosts;
	public double			loggingCostsSp1;
	public double			laborCosts;
	public double			laborCostsSp1;
	public double			expectedNetValue;
	public double			harvestFinesPaid;
	public double			bribesPaid;
	public double			bribeFinesPaid;
	public double			currentLoggedSp1;
	public double			totalLoggedSp1;
	public double			currentLoggedSp2;
	public double	percRisk;
	public int[] 	numTimesSpecies2Logged;
	public int[]	numTimesSpecies2Fined;
	public int 	sumNumTimesSpecies2Logged;
	public int	sumNumTimesSpecies2Fined;
	public int   sumNumTimesBribed;
	public int	sumNumTimesBribeCaught;
	public int[] 		numTimesBribed;
	public int[]		numTimesBribeCaught;
	public double 		m;
	public boolean		havePlaceForNewRoad;
	public int			currentNewRoads;

	// an constructor
	public Logger ( ) {
		id = nextId++;
		live = true;
		netWorth = 0;
		loggingRevenue = 0;
		loggingRevenueSp1 = 0;
		loggingCosts = 0;
		loggingCostsSp1 = 0;
		laborCosts = 0;
		laborCostsSp1 = 0;
		expectedNetValue = 0;
		bribeFinesPaid = 0;
		bribesPaid = 0;
		percRisk = 0;
		currentLoggedSp1 = 0;
		totalLoggedSp1 = 0;
		currentLoggedSp2 = 0;
		numTimesBribed = new int[model.riskMemory] ;
		numTimesBribeCaught = new int[model.riskMemory];
		numTimesSpecies2Fined = new int[model.riskMemory] ;
		numTimesSpecies2Logged = new int[model.riskMemory];
		for(int i = 0; i < model.riskMemory; ++i) {
			numTimesBribed[i] = 0;
			numTimesBribeCaught[i] = 0;
			numTimesSpecies2Logged[i] = 0;
			numTimesSpecies2Fined[i] = 0;
		}
		numTimesBribed[model.riskMemory-1] = 1;  //so that getting caught on the first bribe won't keep them from doing it
		numTimesSpecies2Logged[model.riskMemory-1] = 1;
		m = 0;

		loggingSitesX = new Vector<Integer>();
		loggingSitesY = new Vector<Integer>();
		loggingSitesNetValue = new Vector<Double>();
		loggingSites = new Vector<Observation> ();
		newRoadSites = new Vector<NewRoadObs> ();
		canSeeGrid = new Object2DGrid(model.sizeX, model.sizeY);
		observedCells = new Object2DGrid(model.sizeX, model.sizeY);
		newRoadValues = new Object2DGrid(model.sizeX, model.sizeY);
		cutSp1OrBoth = new Object2DGrid(model.sizeX, model.sizeY);
		for(int i = 0; i < model.sizeX; i++) {
			for(int j = 0; j < model.sizeY; j++) {
				canSeeGrid.putValueAt(i, j, 0);
				observedCells.putValueAt(i, j, 0);
				newRoadValues.putValueAt(i, j, 0);
				cutSp1OrBoth.putValueAt(i, j, 0);
			}
		}


	}


	//initial survey
	public void initialSurvey () {
		//perform the initial survey
		int cellsSeen = 0;
		while(cellsSeen < model.sizeX*model.sizeY*model.initLogSurveyFrac) {

			int newX = ModelParameters.getUniformIntFromTo(0, model.sizeX - 1);
			int newY = ModelParameters.getUniformIntFromTo(0, model.sizeY - 1);
			if(observedCells.getValueAt(newX, newY) == 0) {
				//if they are cells we haven't seen yet, add them to the list
				Observation newObs = new Observation(newX,newY);
				loggingSites.add(newObs);
				observedCells.putValueAt(newX, newY, 1);
				canSeeGrid.putValueAt(newX, newY, newObs.netValue);
				if(newObs.flag0Sp1Only1Both) {
					cutSp1OrBoth.putValueAt(newX, newY,  1000);
				} else {
					cutSp1OrBoth.putValueAt(newX, newY,  0);
				}
				cellsSeen++;
			}
		}

		//order them from highest to lowest mass of species 2
		//Collections.sort( loggingSites, (java.util.Comparator<? super Observation>) new ObsNetValueComparator() );
		removeDuplicatesSp1();
	}
	////////////////////////////////////////////////////////////////////////////
	// step
	// Top level definition of what the agent can do each time its activated.

	public void step () {
		if ( model.getRDebug() > 0 ) 
			System.err.printf( "   -->NGO-step() for id=%d.\n", id );

		recalculateRisk();

		//lookForest ();
		buildNecessaryRoads();
		//evaluateNewRoads(); //roadbuilding based on opening up new access
		logTrees ();
		trimLoggingSitesMemory();



		if ( model.getRDebug() > 1 ) 
			System.err.printf("    <--NGO-step() done.\n" );
	}


	////////////////////////////////////////////////////////////
	//revisit our risk perceptions
	public void recalculateRisk() {

		//add up the number of times that the logger has cut trees he wasn't meant to, and how many times he got caught
		sumNumTimesSpecies2Logged = numTimesSpecies2Logged[0];
		sumNumTimesSpecies2Fined = numTimesSpecies2Fined[0];
		for(int i = 1; i < model.riskMemory; ++i) {
			sumNumTimesSpecies2Logged += numTimesSpecies2Logged[i];
			sumNumTimesSpecies2Fined += numTimesSpecies2Fined[i];
			numTimesSpecies2Fined[i-1] = numTimesSpecies2Fined[i];
			numTimesSpecies2Logged[i-1] = numTimesSpecies2Logged[i];
		}
		numTimesSpecies2Logged[model.riskMemory-1] = 0;
		numTimesSpecies2Fined[model.riskMemory-1] = 0;

		//the risk is simply the ratio of times caught to times committed
		if (sumNumTimesSpecies2Logged > 0) {
			percRisk = (double) sumNumTimesSpecies2Fined / sumNumTimesSpecies2Logged;
		} else {
			percRisk = 0;
			numTimesSpecies2Logged[model.riskMemory-1] = 1;
		}

		//calculate risk of being caught for bribing in the same way
		sumNumTimesBribed = numTimesBribed[0];
		sumNumTimesBribeCaught = numTimesBribeCaught[0];
		for(int i = 1; i < model.riskMemory; ++i) {
			sumNumTimesBribed += numTimesBribed[i];
			sumNumTimesBribeCaught += numTimesBribeCaught[i];
			numTimesBribed[i-1] = numTimesBribed[i];
			numTimesBribeCaught[i-1] = numTimesBribeCaught[i];
		}
		numTimesBribed[model.riskMemory-1] = 0;
		numTimesBribeCaught[model.riskMemory-1] = 0;

		if(sumNumTimesBribed > 0) {
			m = (double) sumNumTimesBribeCaught/sumNumTimesBribed;
		} else {
			m = 0;
			numTimesBribed[model.riskMemory-1] = 1;  //so that getting caught on the first bribe won't keep them from doing it
		}

		//this should be commented if logger is learning risk on his own - this just gives him perfect information of logging risk
		//percRisk = model.monRandCells
	}

	////////////////////////////////////////////////////////////
	//look at the forest to see where we should deforest
	public double calcAccessVolumeSp1() {

		double accessVolumeSp1 = 0;
		double accessVolumeSp2 = 0;
		double accessVolume = 0;
		//make sure the net values of the logging sites are up to date, since with new roads built the value may change

		for ( Observation a : loggingSites ) { 
			if(model.roadNetwork.distanceToRoads.getValueAt(a.x, a.y) <= model.maxDistRoad && a.netValue > 0) { //if we can get to it, and if it's worth getting to...
				accessVolume += a.massSpecies1;
			}
		}


		return accessVolume;

	}
	
	////////////////////////////////////////////////////////////
	//look at the forest to see where we should deforest
	public double calcAccessVolume() {

		double accessVolumeSp1 = 0;
		double accessVolumeSp2 = 0;
		double accessVolume = 0;
		//make sure the net values of the logging sites are up to date, since with new roads built the value may change

		for ( Observation a : loggingSites ) { 
			if(model.roadNetwork.distanceToRoads.getValueAt(a.x, a.y) <= model.maxDistRoad && a.netValue > 0) { //if we can get to it, and if it's worth getting to...
				accessVolumeSp1 += a.massSpecies1;
				accessVolume += a.massSpecies1;
				if(a.flag0Sp1Only1Both) { // model.priceSpecies1PerM3 > model.fineSpecies2PerM3 * percRisk) {
					accessVolumeSp2 += a.massSpecies2;
					accessVolume += a.massSpecies2;
				}
			}
		}


		return accessVolume;

	}
	////////////////////////////////////////////////////////////
	//look at the forest to see where we should deforest
	public void lookForest() {

		//make sure the net values of the logging sites are up to date, since with new roads built the value may change

		for ( Observation a : loggingSites ) { 
			a.updateObservationValue();
			canSeeGrid.putValueAt(a.x, a.y, a.netValue);
			if(a.flag0Sp1Only1Both) {
				cutSp1OrBoth.putValueAt(a.x, a.y,  1000);
			} else {
				cutSp1OrBoth.putValueAt(a.x, a.y,  0);
			}
		}

		//number of new cells to survey
		int numSeeCells =  (int) model.subsLogSurveyFrac * model.sizeX * model.sizeY;

		//randomly select a subset of cells to look at
		for(int i = 0; i < numSeeCells; ++i) {

			int newX = ModelParameters.getUniformIntFromTo(0, model.sizeX - 1);
			int newY = ModelParameters.getUniformIntFromTo(0, model.sizeY - 1);
			if(observedCells.getValueAt(newX, newY) == 0) {
				//if they are cells we haven't seen yet, add them to the list
				Observation newObs = new Observation(newX,newY);
				loggingSites.add(newObs);
				observedCells.putValueAt(newX, newY, 1);
				canSeeGrid.putValueAt(newX, newY, newObs.netValue);
				if(newObs.flag0Sp1Only1Both) {
					cutSp1OrBoth.putValueAt(newX, newY,  1000);
				} else {
					cutSp1OrBoth.putValueAt(newX, newY,  0);
				}
			}
		}

		//order them from highest to lowest mass of species 1
		//Collections.sort( loggingSites, (java.util.Comparator<? super Observation>) new ObsNetValueComparator() );
		removeDuplicatesSp1();

	}

	//THIS IS THE 'HOTSPOT'-based approach to roadbuilding
	//build roads if necessary to log the areas planned for this timestep
	public void buildNecessaryRoads () {
		Vector<Integer> tempRoadX = new Vector<Integer> (0);
		Vector<Integer> tempRoadY = new Vector<Integer> (0);
		Vector<Integer> tempRoadSource = new Vector<Integer> (0);
		
		int[] neighborsX = new int[] {0,0,1,0,-1};
		int[] neighborsY = new int[] {0,1,0,-1,0};
		
		double currentAccessVolume = calcAccessVolumeSp1();
		int i = 0;
		//figure out if this patch is tapped out...
		Vector<Double> maxDistanceVec = model.roadNetwork.distanceToRoads.findMaximum(model.sizeX/2, model.sizeY/2, model.sizeX/2, true, Discrete2DSpace.MOORE);
		Double maxDistance = maxDistanceVec.get(0);
		boolean stillGoodSites = true;

		removeDuplicatesSp1();
		
		while(currentAccessVolume < model.maxLogM3 && maxDistance > model.maxDistRoad && stillGoodSites) { //start with the highest net value cells, and move down


			int currentX = loggingSites.get(i).x;
			int currentY = loggingSites.get(i).y;
			double currentNetValue = loggingSites.get(i).netValue;
			if(currentNetValue <= 0) {
				//we are at the end of good sites
				stillGoodSites = false;
			}
			int dist = (int) model.roadNetwork.distanceToRoads.getValueAt(currentX, currentY);
			int distFromSite = 0;		
			while (dist > 0) {
				
				boolean notFound = true;
				int tempX = currentX;
				int tempY = currentY;
				int times = 1;
				double tempDist = 0;
				int source = -1;
				
				while (notFound) { //find a direction that reduces the distance to another road
					
					int direction = Model.getUniformIntFromTo( 1, 4);  // 1: x-1; 2: y+1; 3: x+1; 4: y-1
					int dx = 0;
					int dy = 0;
					switch (direction) {
					case 1:	dx = -1; source = 0; break;
					case 2: dy = 1; source = 1; break;
					case 3: dx = +1; source = 2; break;
					case 4: dy = -1; source = 3; break;
					}

					tempX = currentX + dx;
					tempY = currentY + dy;

					if( tempX < model.sizeX && tempX >= 0 && tempY < model.sizeY && tempY >= 0) {
						tempDist = model.roadNetwork.distanceToRoads.getValueAt(tempX, tempY);
						if(tempDist < dist) {
							notFound = false;
						}
					}
					times++;
				}
					
				
				distFromSite++;
    			//lay the road, pay the cost
				
				if (distFromSite > model.maxDistRoad && dist > 0) {  //this starts where we first start building roads
					tempRoadX.add(0,currentX);
					tempRoadY.add(0,currentY);
					tempRoadSource.add(0,source); //note, this is the source for the PREVIOUS ROAD NODE
				}
				
    			currentX = tempX;
    			currentY = tempY;
    			//logger can see all areas around road as he/she builds
    			for(int j = 0; j < 5; ++j) {
    				int newTempX = currentX+neighborsX[j];
    				int newTempY = currentY+neighborsY[j];
    				if( newTempX < model.sizeX && newTempX >= 0 && newTempY < model.sizeY && newTempY >= 0) {
    					Observation newObs = new Observation(newTempX,newTempY);
    					loggingSites.add(newObs);
    					observedCells.putValueAt(newTempX, newTempY, 1);
    					canSeeGrid.putValueAt(newTempX, newTempY, newObs.netValue);
    				}
    			}
    			
    			//update the distance we are reducing
    			dist = (int) model.roadNetwork.distanceToRoads.getValueAt(tempX, tempY);
			}
			
			//now go backward along that list and build the road out to the destination (we do it this way
			//so that we can know the distance along roads (for the purposes of road maintenance)
			
			
		 	while(tempRoadX.size() > 0) {
		 		currentX = tempRoadX.get(0);
		 		currentY = tempRoadY.get(0);
		 		int source = tempRoadSource.get(0);
		 		model.roadNetwork.buildRoad(currentX, currentY, 1, source);
		 		//model.roadNetwork.updateDistance(currentX, currentY);
		 		
		 		//logger can see all areas around road as he/she builds
				for(int j = 0; j < 5; ++j) {
					int newTempX = currentX+neighborsX[j];
					int newTempY = currentY+neighborsY[j];
					if( newTempX < model.sizeX && newTempX >= 0 && newTempY < model.sizeY && newTempY >= 0) {
						Observation newObs = new Observation(newTempX,newTempY);
						loggingSites.add(newObs);
						observedCells.putValueAt(newTempX, newTempY, 1);
						canSeeGrid.putValueAt(newTempX, newTempY, newObs.netValue);
						if(newObs.flag0Sp1Only1Both) {
							cutSp1OrBoth.putValueAt(newTempX, newTempY,  1000);
						} else {
							cutSp1OrBoth.putValueAt(newTempX, newTempY,  0);
						}
					}
				}

				//laying the road probably means cutting down species 2.  logger can sell what he cuts for roads,
				//but he also runs the risk of getting fined for having cut...
				double currentS2 = model.ecosystem.getTree(2).getValueAt(currentX, currentY);
				double currentS1 = model.ecosystem.getTree(1).getValueAt(currentX, currentY);
				totalLoggedSp1 += currentS1;
				netWorth = netWorth + (currentS2 * model.priceSpecies2PerM3 + model.priceSpecies1PerM3 * currentS1) * model.fracCutMinorRoad;
				loggingRevenue += (currentS2 * model.priceSpecies2PerM3 + model.priceSpecies1PerM3 * currentS1) * model.fracCutMinorRoad;
				loggingRevenueSp1 += (currentS2 * model.priceSpecies2PerM3 + model.priceSpecies1PerM3 * currentS1) * model.fracCutMinorRoad;
				if(!model.allowedCutRoads) {
					numTimesSpecies2Logged[model.riskMemory-1]++;
				}
				
				netWorth = netWorth - model.costLayMinorRoad;
				loggingCosts += model.costLayMinorRoad;
				loggingCostsSp1 += model.costLayMinorRoad;
				expectedNetValue -= model.costLayMinorRoad;
				
				tempRoadX.removeElementAt(0);
				tempRoadY.removeElementAt(0);
				tempRoadSource.removeElementAt(0);
			}
			
			i++;
			model.roadNetwork.evalDistance();
			currentAccessVolume = calcAccessVolumeSp1();
		}
		removeDuplicatesSp1();
		calcBranchPoints();
	}
	
	
	//look where new roads will be the most profitable
	public void calcBranchPoints () {

		model.roadNetwork.branchPoints = 0;
		int[] neighborsX = new int[] {0,0,1,0,-1};
		int[] neighborsY = new int[] {0,1,0,-1,0};

		int numRoadNodes = model.roadNetwork.roadListX.size()-1; // look at every road cell in the grid as a potential starting point
		double valueSpecies2  = Math.max(0, model.priceSpecies2PerM3- model.fineSpecies2PerM3 * percRisk); //we will need this to calculate total value
		for(int i = 0; i < numRoadNodes; i++) { //for every road cell
			int neighborCount = 0;
			for(int j = 0; j < 5; ++j) { //for each of the moore neighbors of the road cell
				int currentX = model.roadNetwork.roadListX.get(i)+neighborsX[j];
				int currentY = model.roadNetwork.roadListY.get(i)+neighborsY[j];

				double isAlreadyRoad = 99;
				if( currentX < model.sizeX && currentX >= 0 && currentY < model.sizeY && currentY >= 0) {
					isAlreadyRoad = model.roadNetwork.roads.getValueAt(currentX, currentY);
				}


				if(isAlreadyRoad > 0) {
					neighborCount++;
				}

			}
			if(neighborCount > 3) {
				//branch point (includes more than one in, one out, and itself)
				model.roadNetwork.branchPoints++;
			}
		}



	}
	//cut down the trees that we can in this timestep
	public void logTrees () {
		havePlaceForNewRoad = true;
		currentNewRoads = 0;
		
		double volumeToLog = model.maxLogM3;
		int i = 0;
		currentLoggedSp1 = 0;
		currentLoggedSp2 = 0;
		
		removeDuplicates();
		
		while(volumeToLog > 0 && i < loggingSites.size()) {	
			
			if (loggingSites.size() > 7000) {
				int poo = 1;
			}
			int currentX = loggingSites.get(i).x;
			int currentY = loggingSites.get(i).y;
			double currentNetValue = loggingSites.get(i).netValue;
			if(currentNetValue <= 0) {
				//we are at the end of accessible valuable sites - try to build a new road segment
				volumeToLog = 0; //we are at the end of valuable sites

			} else {
				if(model.roadNetwork.distanceToRoads.getValueAt(currentX, currentY) <= model.maxDistRoad) {
					double densityTreeSpecies1 = model.ecosystem.getTree(1).getValueAt(currentX, currentY);
					double densityTreeSpecies2 =  model.ecosystem.getTree(2).getValueAt(currentX, currentY);

					double massSpecies1 = model.ecosystem.getTree(1).getValueAt(currentX, currentY);
					double massSpecies2 = model.ecosystem.getTree(2).getValueAt(currentX, currentY);	

					double deltaRSp1 = 0;
					double deltaRSp2 = 0;

					if(model.loggerList.get(0).currentLoggedSp1 + model.loggerList.get(0).currentLoggedSp2 > 0) {
						deltaRSp1 = massSpecies1/(model.loggerList.get(0).currentLoggedSp1 + model.loggerList.get(0).currentLoggedSp2 );
						deltaRSp2 = massSpecies2/(model.loggerList.get(0).currentLoggedSp1 + model.loggerList.get(0).currentLoggedSp2 + massSpecies1);
					}

					double changeCostsSp1 = deltaRSp1*model.maxProbCheckpointCatch*(model.loggerList.get(0).currentLoggedSp2)*model.fineSpecies2PerM3;
					double changeCostsSp2 = -deltaRSp2*model.maxProbCheckpointCatch*(model.loggerList.get(0).currentLoggedSp2 + massSpecies2)*model.fineSpecies2PerM3;

					double valueSp2 = (model.priceSpecies2PerM3 - model.fineSpecies2PerM3 * percRisk)*massSpecies2 + changeCostsSp2;

					//have to recalculate the net returns to include the change in probability of a checkpoint catch.  since the change costs
					//change as the truck fills up, we have to redo this...
					double netReturnSp1 = ((loggingSites.get(i).netReturnSp1M3 * massSpecies1) + changeCostsSp1) / massSpecies1;
					if(massSpecies1 == 0) {
						netReturnSp1 = 0;
					}
					double netReturnSp1Sp2 = ((loggingSites.get(i).netReturnSp1Sp2M3 * (massSpecies1 + massSpecies2)) + changeCostsSp1 + changeCostsSp2) / (massSpecies1 + massSpecies2);
					
					double netReturnSp2 = valueSp2 / massSpecies2;
					
					if(netReturnSp1 == 0) {
						int poo = 1;
					}
					//if (densityTreeSpecies1 > 0) {  //ignore it if it is just a blank one on the list
					if(densityTreeSpecies1 < volumeToLog) {
						model.ecosystem.getTree(1).putValueAt(currentX, currentY, 0);
						netWorth = netWorth + model.priceSpecies1PerM3 * densityTreeSpecies1;
						loggingRevenue += model.priceSpecies1PerM3 * densityTreeSpecies1;
						loggingRevenueSp1 += model.priceSpecies1PerM3 * densityTreeSpecies1;
						volumeToLog -= densityTreeSpecies1;
						currentLoggedSp1 += densityTreeSpecies1;
						totalLoggedSp1 += densityTreeSpecies1;
					}else {
						model.ecosystem.getTree(1).putValueAt(currentX, currentY, densityTreeSpecies1 - volumeToLog);
						netWorth = netWorth + model.priceSpecies1PerM3 * model.maxHarvestOnePeriod;
						loggingRevenue += model.priceSpecies1PerM3 * model.maxHarvestOnePeriod;
						loggingRevenueSp1 += model.priceSpecies1PerM3 * model.maxHarvestOnePeriod;
						currentLoggedSp1 += volumeToLog;
						totalLoggedSp1 += volumeToLog;
						volumeToLog = 0;
					}


					// if the net pay off of cutting illegally is positive, cut species 2
					double aveReturnSp1 = (loggingRevenueSp1 - laborCostsSp1 - loggingCostsSp1)/ totalLoggedSp1;
					if (totalLoggedSp1 == 0) {
						aveReturnSp1 = 0;
					}
					if(netReturnSp2 > aveReturnSp1) {
					//if(netReturnSp1Sp2 > netReturnSp1) {
					//if(loggingSites.get(i).flag0Sp1Only1Both) { //THIS NEEDS TO BE UPDATED BEFORE CHECKPOINT MONITORING WILL WORK PROPERLY!!!!!
						//if(valueSp2 > 0) {
						if(densityTreeSpecies2 < volumeToLog) {
							model.ecosystem.getTree(2).putValueAt(currentX, currentY, 0);
							netWorth = netWorth + model.priceSpecies2PerM3 * densityTreeSpecies2;
							loggingRevenue += model.priceSpecies2PerM3 * densityTreeSpecies2;
							volumeToLog -= densityTreeSpecies2;
							currentLoggedSp2 += densityTreeSpecies2;
						}else {
							model.ecosystem.getTree(2).putValueAt(currentX, currentY, densityTreeSpecies2 - volumeToLog);
							netWorth = netWorth + model.priceSpecies2PerM3 * volumeToLog;
							loggingRevenue += model.priceSpecies2PerM3 * volumeToLog;
							currentLoggedSp2 += volumeToLog;
							volumeToLog = 0;
						}	
						numTimesSpecies2Logged[model.riskMemory-1]++;
					}


					expectedNetValue += currentNetValue;

					loggingSites.get(i).updateObservation();
					if(loggingSites.get(i).massSpecies1 == 0) {
						loggingSites.remove(i); //ie, we don't want to come back to this one, since we already were here and decided not to cut sp2
					} else {
						i++;
					}
					//cellsToLog--;

					//logger can see all areas around site as he/she logs (but we won't reorder the list just yet)
					int[] neighborsX = new int[] {0,0,1,0,-1};
					int[] neighborsY = new int[] {0,1,0,-1,0};
					for(int j = 0; j < 5; ++j) {
						int newTempX = currentX+neighborsX[j];
						int newTempY = currentY+neighborsY[j];
						if( newTempX < model.sizeX && newTempX >= 0 && newTempY < model.sizeY && newTempY >= 0) {
							Observation newObs = new Observation(newTempX,newTempY);
							if(newObs.massSpecies1 > 0) { //ie, we HAVENT already been there...
								loggingSites.add(newObs);
								observedCells.putValueAt(newTempX, newTempY, 1);
								canSeeGrid.putValueAt(newTempX, newTempY, newObs.netValue);
								if(newObs.flag0Sp1Only1Both) {
									cutSp1OrBoth.putValueAt(newTempX, newTempY,  1000);
								} else {
									cutSp1OrBoth.putValueAt(newTempX, newTempY,  0);
								}
							}
						}
					}
					
					//pay the maintenance costs to get to this point.

					int nearestRoadX = model.roadNetwork.roadListX.get((int) model.roadNetwork.closestRoadNode.getValueAt(currentX, currentY));
					int nearestRoadY = model.roadNetwork.roadListY.get((int) model.roadNetwork.closestRoadNode.getValueAt(currentX, currentY));
					double currentMaintCosts = model.roadNetwork.cumMaintenanceCostToAccess.getValueAt(nearestRoadX, nearestRoadY);
					if (currentMaintCosts > 0) {
						netWorth -= model.roadNetwork.cumMaintenanceCostToAccess.getValueAt(nearestRoadX, nearestRoadY);
						loggingCostsSp1 -= model.roadNetwork.cumMaintenanceCostToAccess.getValueAt(nearestRoadX, nearestRoadY);
						model.roadNetwork.maintainRoad(nearestRoadX, nearestRoadY);

						for ( Observation a : loggingSites ) { //adjust the net value since there will be cost changes...
							a.updateObservationValue();
							canSeeGrid.putValueAt(a.x, a.y, a.netValue);
							if(a.flag0Sp1Only1Both) {
								cutSp1OrBoth.putValueAt(a.x, a.y,  1000);
							} else {
								cutSp1OrBoth.putValueAt(a.x, a.y,  0);
							}
						}

						//reorder the sites since this will have changed their relative value
						removeDuplicates();
						//Collections.sort( loggingSites, (java.util.Comparator<? super Observation>) new ObsNetValueComparator() );

						//go back to the beginning of the list since we have re-ordered it
						i = 0;
					}
				} else {
					i++;
				}

				
				//removeDuplicates(); 

			}

		}
		netWorth -= model.costHarvestPerM3 * (currentLoggedSp1 + currentLoggedSp2);
		loggingCosts += model.costHarvestPerM3 * (currentLoggedSp1 + currentLoggedSp2);
		laborCosts += model.costHarvestPerM3 * (currentLoggedSp1 + currentLoggedSp2);
		laborCostsSp1 += model.costHarvestPerM3 * (currentLoggedSp1);
		//}
		removeDuplicates();
	}

	//sort, and remove duplicates from the vector of memory. option to also remove netValues less than 0, by just switching out the other if statement
	public void removeDuplicates () {
		int i = 0;
		Collections.sort( loggingSites, (java.util.Comparator<? super Observation>) new ObsIDComparator() );
		while (i < loggingSites.size()-1) {
			//if((loggingSites.get(i).x == loggingSites.get(i+1).x && loggingSites.get(i).y == loggingSites.get(i+1).y) || loggingSites.get(i+1).netValue <= 0) {
			double currentDistRoad = model.roadNetwork.distanceToRoads.getValueAt(loggingSites.get(i).x, loggingSites.get(i).y);
			if((loggingSites.get(i).elementID == loggingSites.get(i+1).elementID) || loggingSites.get(i).massSpecies1 == 0 || ((loggingSites.get(i).netValue < 0) && currentDistRoad <= model.maxDistRoad)) {	
				loggingSites.removeElementAt(i);
			} else {
				i = i + 1;
			}
		}		
		Collections.sort( loggingSites, (java.util.Comparator<? super Observation>) new ObsNetValueComparator() );

	}
	
	//sort, and remove duplicates from the vector of memory. option to also remove netValues less than 0, by just switching out the other if statement
	public void removeDuplicatesSp1 () {
		int i = 0;
		Collections.sort( loggingSites, (java.util.Comparator<? super Observation>) new ObsIDComparator() );
		while (i < loggingSites.size()-1) {
			//if((loggingSites.get(i).x == loggingSites.get(i+1).x && loggingSites.get(i).y == loggingSites.get(i+1).y) || loggingSites.get(i+1).netValue <= 0) {
			double currentDistRoad = model.roadNetwork.distanceToRoads.getValueAt(loggingSites.get(i).x, loggingSites.get(i).y);
			if((loggingSites.get(i).elementID == loggingSites.get(i+1).elementID) || loggingSites.get(i).massSpecies1 == 0 || ((loggingSites.get(i).netValue < 0) && currentDistRoad <= model.maxDistRoad)) {	
				loggingSites.removeElementAt(i);
			} else {
				i = i + 1;
			}
		}		
		Collections.sort( loggingSites, (java.util.Comparator<? super Observation>) new ObsSp1Comparator() );

	}
	

	//sort, and remove duplicates from the vector of new road sites. also remove potentialValues less than 0
	public void removeDuplicatesRoads () {
		int i = 0;
		Collections.sort( newRoadSites, (java.util.Comparator<? super NewRoadObs>) new newRoadNodeIDComparator() );
		while (i < newRoadSites.size()-1) {
			if((newRoadSites.get(i).elementID == newRoadSites.get(i+1).elementID) || newRoadSites.get(i+1).potentialValue <= 0) {
				newRoadSites.removeElementAt(i+1);
			} else {
				i = i + 1;
			}
		}
		Collections.sort( newRoadSites, (java.util.Comparator<? super NewRoadObs>) new newRoadNodeComparator() );
		if(newRoadSites.get(0).potentialValue <= 0) {
			newRoadSites.removeElementAt(0);  //just check that last one...
		}
	}
	//trim down the vector of memory
	public void trimLoggingSitesMemory () {
		while (loggingSites.size() > model.lengthMemory) {
			int currentSize = loggingSites.size()-1;
			int currentX = loggingSites.get(currentSize).x;
			int currentY = loggingSites.get(currentSize).y;
			observedCells.putValueAt(currentX, currentY, 0);
			canSeeGrid.putValueAt(currentX, currentY, 0);
			cutSp1OrBoth.putValueAt(currentX, currentY, 0);
			loggingSites.removeElementAt(currentSize);

		}
	}

	private class ObsSp1Comparator implements Comparator {
		public final int compare ( Object obs1, Object obs2 ) {
			double d1 = ( (Observation) obs1 ).massSpecies1;
			double d2 = ( (Observation) obs2 ).massSpecies1;
			if ( d1 < d2 ) //orders values descending...
				return 1;
			else if ( d1 > d2 )
				return -1;
			return 0;
		}
	}
	
	private class ObsNetValueComparator implements Comparator {
		public final int compare ( Object obs1, Object obs2 ) {
			double d1 = ( (Observation) obs1 ).netValue;
			double d2 = ( (Observation) obs2 ).netValue;
			if ( d1 < d2 ) //orders values descending...
				return 1;
			else if ( d1 > d2 )
				return -1;
			return 0;
		}
	}
	

	private class ObsIDComparator implements Comparator {
		public final int compare ( Object obs1, Object obs2 ) {
			double d1 = ( (Observation) obs1 ).elementID;
			double d2 = ( (Observation) obs2 ).elementID;
			if ( d1 < d2 ) //orders values descending...
				return 1;
			else if ( d1 > d2 )
				return -1;
			return 0;
		}
	}


	private class newRoadNodeComparator implements Comparator {
		public final int compare ( Object newR1, Object newR2 ) {
			double d1 = ( (NewRoadObs) newR1 ).potentialValue;
			double d2 = ( (NewRoadObs) newR2 ).potentialValue;
			if ( d1 < d2 ) //orders values descending...
				return 1;
			else if ( d1 > d2 )
				return -1;
			return 0;
		}
	}

	private class newRoadNodeIDComparator implements Comparator {
		public final int compare ( Object newR1, Object newR2 ) {
			double d1 = ( (NewRoadObs) newR1 ).elementID;
			double d2 = ( (NewRoadObs) newR2 ).elementID;
			if ( d1 < d2 ) //orders values descending...
				return 1;
			else if ( d1 > d2 )
				return -1;
			return 0;
		}
	}

	////////////////////////////////////////////////////////////
	// printOut
	// print fields to System.out.
	//
	// Note: we can use the class's field variables in all methods.

	public void printOut ( ) {
		System.out.printf( " - NGO id %2d; live=%b).\n",
				id, live );
	}

	////////////////////////////////////////////////////////////////////////////
	// setters and getters
	//
	public int getId() {  return id; }

}
