package CAFI;

/**
Ministry.java 

Class for Ministry, a type of Agent

**/

import java.util.Collections;
import java.util.Comparator;
import java.util.Formatter;
import java.util.Vector;
import java.awt.Point;
import java.util.ArrayList;

import uchicago.src.sim.util.*;
import uchicago.src.sim.gui.*;
import uchicago.src.sim.space.*;
import uchicago.src.sim.gui.ColorMap;
import java.awt.Color;

public class Ministry extends Agent {
	// "class" variables -- one value for all instances 
    public  static int          	nextId = 0; // to give each a unique id

	// fields (instance variables)  
	int 				id;		// unique id number for each instance
	public Object2DGrid NGOReports;
	public Object2DGrid monitorReports;



	// a constructor
	public Ministry ( ) {
		id = nextId++;
		
		NGOReports = new Object2DGrid(model.sizeX, model.sizeY);
		monitorReports = new Object2DGrid(model.sizeX, model.sizeY);
		for(int i = 0; i < model.sizeX; i++) {
			for(int j = 0; j < model.sizeY; j++) {
				NGOReports.putValueAt(i, j, 0);	
				monitorReports.putValueAt(i, j, 0);	
			}
		}
		
	}

	////////////////////////////////////////////////////////////////////////////
	// step
	// Top level definition of what the agent can do each time its activated.

	public void step () {
		if ( model.getRDebug() > 0 ) 
			System.err.printf( "   -->Ministry-step() for id=%d.\n", id );

		//!!!right now this code still hardwired for ONE NGO, ONE enforcer; needs to be updated when we move on from here
		NGO currentNGO = model.ngoList.get(0);
		LocalEnforcer currentMonitor = model.localEnforcersList.get(0);
		Logger currentLogger = model.loggerList.get(0);
		
		//get reports from NGOs 
		while (currentNGO.reportSites.size() > 0) {
			Report currentReport = currentNGO.reportSites.get(0);
			NGOReports.putValueAt(currentReport.getX(), currentReport.getY(), currentReport.getMassLoggedIllegally());
			currentNGO.reportSites.remove(0);
		}
		
		//get reports from monitors and compare
		while (currentMonitor.reportSites.size() > 0) {
			Report currentReport = currentMonitor.reportSites.get(0);
			int currentX = currentReport.getX();
			int currentY = currentReport.getY();
			
			monitorReports.putValueAt(currentX, currentY, currentReport.getMassLoggedIllegally());
			
			double currentMonitorReport = monitorReports.getValueAt(currentX, currentY);
			double currentNGOReport = NGOReports.getValueAt(currentX, currentY);
			
			if(model.bribeOn && currentMonitorReport < currentNGOReport) {
				//monitor has been caught in a bribe!!!
				double revenueIllegal = (currentNGOReport - currentMonitorReport) * model.priceSpecies2PerM3;
				currentMonitor.bribeFinesPaid += model.phiMonitor*revenueIllegal;
				currentLogger.netWorth -= model.phiLogger*revenueIllegal;
				currentLogger.bribeFinesPaid += model.phiLogger*revenueIllegal;
				currentMonitor.numTimesBribeCaught[model.riskMemory-1]++;
				currentLogger.numTimesBribeCaught[model.riskMemory-1]++;
			}
			currentMonitor.reportSites.remove(0);
		}
		
		
		if ( model.getRDebug() > 1 ) 
			System.err.printf("    <--Ministry-step() done.\n" );
	}
	

	
	////////////////////////////////////////////////////////////
	// printOut
	// print fields to System.out.
	//
	// Note: we can use the class's field variables in all methods.

	public void printOut ( ) {
		System.out.printf( " - LocalEnforcer id %2d (x,y=%d,%d; live=%b).\n",
						   id );
	}


}
