package CAFI;

// First model of loggers, NGO's, communities, etc.
//
//
//

import java.io.*;
import java.math.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Formatter;
import java.util.Vector;
import java.awt.Point;

import java.lang.reflect.*;

import uchicago.src.sim.engine.Schedule;
import uchicago.src.sim.engine.SimModelImpl;
import uchicago.src.sim.util.*;
import uchicago.src.sim.gui.*;
import uchicago.src.sim.space.*;


public class Model extends ModelParameters {

	// fields for run-time parameters
	public double			cellSize; //size in hectares of one cell
	public int 				numLocalEnforcers;  // initial number of LocalEnforcers
    public int              sizeX, sizeY;   // integer size of the world 
    public int				cuttingDays; // number of days in patch across which cutting allowance is spaced
    
    //these are the variables that get used by the model, and are normalized by cellSize
    public double			maintenanceCostPerYearPerMinorRoadCell;
    public double			maintenanceCostPerYearPerMajorRoadCell;
    public double			fracCutMinorRoad;
    public double			fracCutMajorRoad;
	public double			maxVolumeS1;
	public double			maxVolumeS2;
	public double			maxVolumeS0;
	public double			maxCost;
	public int				numSpecies;
	public int				monRandCells;
	public int				NGORandCells;
	public double			initLogSurveyFrac;
	public double			subsLogSurveyFrac;
	public double			costLayMinorRoad;
	public double			costLayMajorRoad;
	public double			costHarvestPerM3;
	public double			maxHarvestOnePeriod;
	public double			priceSpecies1PerM3;
	public double			priceSpecies2PerM3;
	public double			fineSpecies2PerM3;
	public int				maxDistRoad;
	public int				maxMonCells;
	public int				maxNGOCells;
	
	//these are the raw values that get input, and need to be scaled before they can be used
    public double			maintenanceCostPerYearPerKmMinorRoad;
    public double			maintenanceCostPerYearPerKmMajorRoad;
    public double			widthMinorRoad;
    public double			widthMajorRoad;
	public double			maxDensityS1;
	public double			maxDensityS2;
	public double			maxDensityS0;
	public double				monRandArea;
	public double				NGORandArea;
	public double			costLayMinorRoadPerKm;
	public double			costLayMajorRoadPerKm;
	public double			maxHarvestOnePeriodM3;
	public double				maxDistRoadMeters;
	public double				maxLogM3;
	public double				maxMonArea;
	public double				maxNGOArea;
	
	public int				lengthMemory;
	public double[]			initialBiomass;
	public double[]			currentBiomass;
	public int				numSteps;
	public int				graphUpdateFrequency;
	public double			maxProbCheckpointCatch;
	public int				kernelRadius;
	public double			kernelWeightParam;
	
	//bribing parameters
	public double			phiLogger;  //per unit rate levied against loggers caught bribing
	public double			phiMonitor; //per unit rate levied against monitors caught bribing
	public double			betaBribe; //bargaining power between monitor and logger --> beta near 0 means monitor strong, vice versa
	public double			sigmaBribe; //relative standard deviation of bribes
	public int			riskMemory; //num timesteps agents remember fining/bribing events
	public boolean			bribeOn;  //are we doing bribing - note, without bribing, NGOs are kinda meaningless
	public boolean			edgeOn;	//have a boundary around the edges equal to maxDistRoad - prevents edge effects
	public boolean			treeDistTestOn; //distribute trees in test pattern - see Tree.java for exact pattern
	public boolean			allowedCutRoads; //allowed to cut whatever trees they want in order to build roads
	public boolean			useKernelSmooth; //smooth the distribution and create spatial auto-correlation
	public boolean			usePowerDist; //use power dist rather than uniform dist
	public double			powerAlpha; //parameter for power distribution
	
	// fields for model "structures"
	public ArrayList<AgentInSpace>   allInSpaceList;
	public ArrayList<LocalEnforcer>  localEnforcersList;
	public ArrayList<NGO>			 ngoList;
	public ArrayList<Logger>		loggerList;
	public Ecosystem				ecosystem;
	public RoadNetwork				roadNetwork;

	public GridWorld   		world;        	  // a discrete world, bounded by default
	public boolean			useTorus = false; // if true, world is torus
	public Ministry			ministry;

	// fields for aggregate measures
	public double			localEnforcersAvgX;  // for testing
	
	//documents
	public PrintWriter 		forestFile;
	public PrintWriter 		roadFile;




	// a field used by repast for keeping track of model steps
	public Schedule   		schedule;		  // repast schedule of events

	/////////////////////////////////////////////////////////////////////////////
	// To add a new parameter that can be set at run time (via command line or gui)
	// do the following steps:
	// 0. Add a description of the parameter in  printProjectHelp() .
	// 1. Add an instance variable (field) for the parameter to Model.java .
	// 2. In the method addModelSpecificParameters
	//    add alias and long name for Model parameters you want to set at run time
	//    the long name should be same as instance variable
	// 3. Add setter and getter methods for the parameter
	// 4. In the method getInitParam, add the parameter name to the array of Strings.
	// 5. In userSetup(), give the parameter a default value.
	// 6. Compile and test by running model and seeing that the parameter:
	//    - is on the GUI panel, in the report file, and settable on the command line,
	//      eg     bin/batchrun.sh newParName=1.4
	// Note: the generic parameters from ModelParameters are already available.
	//       enter    bin/guirun.sh -h  
	// to see all parameters and defaults, including new ones you add.

	public void addModelSpecificParameters () {
		parametersMap.put( "X", "sizeX" );
		parametersMap.put( "Y", "sizeY" );
		parametersMap.put( "cS", "cellSize" );
		parametersMap.put( "tMnC", "maintenanceCostPerYearPerKmMinorRoad");
		parametersMap.put( "tMjC", "maintenanceCostPerYearPerKmMajorRoad");
		parametersMap.put( "iLSF", "initLogSurveyFrac");
		parametersMap.put( "sLSF", "subsLogSurveyFrac");
		parametersMap.put( "mDS1", "maxDensityS1" );
		parametersMap.put( "mDS2", "maxDensityS2" );
		parametersMap.put( "mDS3", "maxDensityS0" );
		parametersMap.put( "mRC", "monRandArea" );
		parametersMap.put( "nRC", "NGORandArea" );
		parametersMap.put( "cLNR", "costLayMinorRoadPerKm" );
		//parametersMap.put( "cLJR", "costLayMajorRoad" );
		parametersMap.put( "pS1M", "priceSpecies1PerM3" );
		parametersMap.put( "pS2M", "priceSpecies2PerM3" );
		parametersMap.put( "fS2M", "fineSpecies2PerM3" );
		parametersMap.put( "mDR", "maxDistRoadMeters" );
		parametersMap.put( "mLC", "maxLogM3" );
		parametersMap.put( "mMC", "maxMonArea" );
		parametersMap.put( "mNC", "maxNGOArea" );
		parametersMap.put( "lM", "lengthMemory" );
		parametersMap.put( "pL", "phiLogger" );
		parametersMap.put( "pM", "phiMonitor" );
		parametersMap.put( "bB", "betaBribe" );
		parametersMap.put( "bO", "bribeOn" );
		parametersMap.put( "aCR", "allowedCutRoads" );
		parametersMap.put( "eO", "edgeOn" );
		parametersMap.put( "sB", "sigmaBribe" );
		parametersMap.put( "rM", "riskMemory" );
		parametersMap.put( "gUF", "graphUpdateFrequency" );
		parametersMap.put( "mPCC", "maxProbCheckpointCatch" );
		parametersMap.put( "uKS", "useKernelSmooth" );
		parametersMap.put( "uPD", "usePowerDist" );
		parametersMap.put( "kR", "kernelRadius" );
		parametersMap.put( "kWP", "kernelWeightParam" );
		parametersMap.put( "pA", "powerAlpha" );
		parametersMap.put( "cD", "cuttingDays" );
		parametersMap.put( "cHM", "costHarvestPerM3" );
	}

	// Specify what appears in the repast parameter panel
	public String[] getInitParam () {
		String[] params = { "sizeX", "sizeY", "cellSize", "maintenanceCostPerYearPerKmMinorRoad", "maintenanceCostPerYearPerKmMajorRoad", "cuttingDays",
				"initLogSurveyFrac", "subsLogSurveyFrac","maxDensityS1","maxDensityS2","maxDensityS0", "bribeOn", "powerAlpha",
				"monRandArea", "costLayMinorRoadPerKm", "priceSpecies1PerM3","priceSpecies2PerM3", "allowedCutRoads", "maxProbCheckpointCatch", "costHarvestPerM3",
				"fineSpecies2PerM3", "maxDistRoadMeters", "phiLogger","phiMonitor", "betaBribe", "sigmaBribe", "riskMemory","kernelWeightParam", "usePowerDist",
				"maxLogM3","maxMonArea","lengthMemory", "maxNGOArea", "NGORandArea", "graphUpdateFrequency", "edgeOn", "treeDistTestOn","useKernelSmooth","kernelRadius",		
				// these are from the super class:
				"rDebug", "seed" };
		return params;
	}

	///////////////////////////////////////////////////////////////////////////
	// userSetup()
	// discard any existing model parts, and re-initialize to defaults
	
	public void userSetup() {
		if ( rDebug > 0 )
			System.out.printf( "==> userSetup...\n" );

		numLocalEnforcers = 1;
		AgentInSpace.setNextId( 0 );
		LocalEnforcer.setNextId( 0 );
		NGO.setNextId( 0 );
		Tree.setNextId( 0 );

		cellSize = (double) 1/4;
		allInSpaceList = new ArrayList<AgentInSpace> ();  	// new empty list 
		localEnforcersList = new ArrayList<LocalEnforcer> ();
		ngoList = new ArrayList<NGO>();
		loggerList = new ArrayList<Logger> ();
		world = null;						// toss out the world
		useTorus = false;
		sizeX = 81;  sizeY = 81;
		maxDensityS1 = 60;
		maxDensityS2 = 15;
		maxDensityS0 = 180;
		maxCost = 3000;
		numSpecies = 3;
		monRandArea = 0;
		NGORandArea = 0;
		costLayMinorRoadPerKm = 10000; 
		costLayMajorRoadPerKm = 10000;
		priceSpecies1PerM3 = 200;
		priceSpecies2PerM3 = 400;
		fineSpecies2PerM3 = 1000;
		maxDistRoadMeters = 300;
		maxLogM3 = 150*365;
		maxMonArea = 0;
		maxNGOArea = 0;
		lengthMemory = 81*81;
		numSteps = 0;
		riskMemory = 800;
		graphUpdateFrequency = 20;
		maintenanceCostPerYearPerKmMinorRoad = 0;
		maintenanceCostPerYearPerKmMajorRoad = 0;
		widthMinorRoad = 10;
		widthMajorRoad = 50;
		initLogSurveyFrac = 1;
		subsLogSurveyFrac = 0;
		maxProbCheckpointCatch = 0;
		kernelRadius = 5;
		kernelWeightParam = 1;
		powerAlpha = 2.5;
		cuttingDays = 100;
		costHarvestPerM3 = 30;
		
		phiLogger = 2;
		phiMonitor = 1;
		betaBribe = 0.5;
		sigmaBribe = .1;
		bribeOn = false;
		edgeOn = false;
		treeDistTestOn = false;
		allowedCutRoads = true;
		useKernelSmooth = true;
		usePowerDist = true;
		
	}

	///////////////////////////////////////////////////////////////////////////
	// userBuildModel
	// create objects that constitute the model:
	// - a 2D world
	// - a list of LocalEnforcer's, placed at random in the world
	
	public void userBuildModel () {
		if ( rDebug > 0 )
			System.out.printf( "==> userBuildModel...\n" );

		//now calculate the scaled variables needed by the model
		double cellLength = Math.sqrt(cellSize*10000); //in meters
	    maintenanceCostPerYearPerMinorRoadCell = (double) maintenanceCostPerYearPerKmMinorRoad/1000*cellLength/cuttingDays;
	    maintenanceCostPerYearPerMajorRoadCell = (double) maintenanceCostPerYearPerKmMajorRoad/1000*cellLength/cuttingDays;
	    fracCutMinorRoad = Math.max(1,(double) widthMinorRoad/cellLength);
	    fracCutMajorRoad = Math.max(1,(double) widthMajorRoad/cellLength);
		maxVolumeS1 = maxDensityS1*cellSize;
		maxVolumeS2 = maxDensityS2*cellSize;
		maxVolumeS0 = maxDensityS0*cellSize;
		monRandCells = (int) Math.round(monRandArea/cellSize);
		NGORandCells = (int) Math.round(NGORandArea/cellSize);
		costLayMinorRoad = costLayMinorRoadPerKm/1000*cellLength;
		costLayMajorRoad = costLayMajorRoadPerKm/1000*cellLength;
		maxDistRoad = (int) Math.round(maxDistRoadMeters/cellLength);
		maxMonCells = (int) Math.round(maxMonArea/cellSize);
		maxNGOCells = (int) Math.round(maxNGOArea/cellSize);
		maxLogM3 = (double) maxLogM3/cuttingDays;
		
		// create the 2D grid world of requested size
		if ( useTorus )
			world = new TorusWorld( sizeX, sizeY );
		else
			world = new GridWorld( sizeX, sizeY );

		// tell the Agent class about this (Model) and world addresses
		// so that they can send messages to them, e.g.,
		// to query the world about cell contents.
		AgentInSpace.setModel( this );
		AgentInSpace.setWorld( world );
		NGO.setModel( this );
		NGO.setWorld( world );  // NGO's not in world, but know of it
		Ecosystem.setModel( this );
		Ecosystem.setWorld( world ); 
		Logger.setModel( this );
		Logger.setWorld( world ); 
		RoadNetwork.setModel( this );
		RoadNetwork.setWorld( world ); 
		Tree.setModel( this );
		Tree.setWorld( world );
		Observation.setModel( this );
		Observation.setWorld( world );
		NewRoadObs.setModel( this );
		NewRoadObs.setWorld( world );
		
		// create various agents 
		createLocalEnforcers();
		createNGOs();
		createEcosystem();
		createRoads();
		createLoggers();

		for ( Logger a : loggerList )  
			a.initialSurvey();
		
		ministry = new Ministry();
		
		initialBiomass = new double[numSpecies];
		currentBiomass = new double[numSpecies];
		for(int i = 0; i < numSpecies; ++i) {
			initialBiomass[i] = ecosystem.speciesList.get(i).totalBiomass;
			currentBiomass[i] = ecosystem.speciesList.get(i).totalBiomass;
		}
		

		
		calcModelStats();			// calc init stats

		if ( rDebug > 0 )
			System.out.printf( "<==  userbuildModel done.\n" );

	}


	////////////////////////////////////////////////////////////////////////////
	// createLocalEnforcers
	// create the LocalEnforcer's, add to world and to the localEnforcerslist
	public void createLocalEnforcers () {
		if ( rDebug > 0 )
			System.out.printf( "==> createLocalEnforcers...\n" );

		for ( int i = 0; i < numLocalEnforcers; ++i ) {
			LocalEnforcer a = new LocalEnforcer();
			localEnforcersList.add( a );
		}

	}


	////////////////////////////////////////////////////////////////////////////
	// createNGOs
	// create the NGOs, add to list of them.
	// note these are aspatial.
	//
	// currently just creates 2 of them.
	// we may want to load these with particular attributes, etc.

	public void createNGOs () {
		if ( rDebug > 0 )
			System.out.printf( "==> createNGOs...\n" );

		for ( int i = 0; i < 1; ++i ) {
			NGO a = new NGO();
			ngoList.add( a );
		}

	}

	////////////////////////////////////////////////////////////////////////////
	// createLoggers
	// create the Loggers, add to list of them.
	// note these are aspatial.
	//

	public void createLoggers () {
		if ( rDebug > 0 )
			System.out.printf( "==> createLoggers...\n" );

		for ( int i = 0; i < 1; ++i ) {
			Logger a = new Logger();
			loggerList.add( a );
		}

	}

	////////////////////////////////////////////////////////////////////////////
	// createEcosystem
	// create the Ecosystem.


	public void createEcosystem () {
		if ( rDebug > 0 )
			System.out.printf( "==> createEcosystem...\n" );

		
			double[] volumeList = new double[numSpecies];
			volumeList[1] = maxVolumeS1;
			volumeList[2] = maxVolumeS2;
			volumeList[0] = maxVolumeS0;
			ecosystem = new Ecosystem();
			for(int i = 0; i < numSpecies; ++i) {
				ecosystem.addSpecies(volumeList[i],i);
			}

	}
	
	
	////////////////////////////////////////////////////////////////////////////
	// createRoads
	// create the Road network.


	public void createRoads () {
		if ( rDebug > 0 )
			System.out.printf( "==> createRoads...\n" );

			roadNetwork = new RoadNetwork();
			//make some initial roads - right now this is just to have something there
			//1 - minor roads
			//2 - major roads
					
			int j = (int) sizeY/3;
			/*for(int i = 0; i < sizeX; i++) {			
				roadNetwork.buildRoad(i, j, 1);
			}
			j = (int) 2*sizeY/3;
			for(int i = 0; i < sizeX; i++) {			
				roadNetwork.buildRoad(i, j, 1);
			}*/
			
			
			
			int i = (int) sizeX/2;
			int k = 0;
			j = 0;
			roadNetwork.buildRoad(i, j, 2,-1);
			roadNetwork.distanceAlongRoads.putValueAt(i, j, k * maintenanceCostPerYearPerMajorRoadCell);
			k++;
			
			roadNetwork.originX = i;
			roadNetwork.originY = j;
			
			for(j = 1; j < sizeY; j++) {			
				roadNetwork.buildRoad(i, j, 2, 3);
				roadNetwork.distanceAlongRoads.putValueAt(i, j, k * maintenanceCostPerYearPerMajorRoadCell);
				k++;
			}	

			//this is to 'seed' the roads when we are using the local optimizing road algorithm evaluateNewRoad
			//don't need it when we are doing hotspots
			//roadNetwork.buildRoad(((int) sizeX/2)-1, maxDistRoad, 1, 2);
			//roadNetwork.distanceAlongRoads.putValueAt(((int) sizeX/2)-1, maxDistRoad, (maxDistRoad) * maintenanceCostPerYearPerMajorRoadCell + maintenanceCostPerYearPerMinorRoadCell);
			//double closestNode = roadNetwork.closestRoadNode.getValueAt(((int) sizeX/2)-1, maxDistRoad);
			for( j = 0; j < sizeY; j++) {
				double closestNode = roadNetwork.closestRoadNode.getValueAt((int) sizeX/2, j);
				for( i = 0; i < sizeX; i++) {
					if(roadNetwork.closestRoadNode.getValueAt(i, j) < 0) {
						roadNetwork.closestRoadNode.putValueAt(i, j, closestNode);	
					}
				}
			}
			roadNetwork.evalDistance();

	}
	
	///////////////////////////////////////////////////////////////////////////
	// step
	//
	// The top of the "conceptual" model's main dynamics
	// This defines what happens each time step (user presses Step or Run buttons...)
	// Currently:  Just tell each agent to execute its own step method.
	//
	// NB:  At the end, it calls stepReport() to write to report file.

	public void step () {
		if ( rDebug > 0 )
			System.out.printf( "==> Model step %.0f:\n", getTickCount() );

		if ( getTickCount() % cuttingDays == 1  ) {
					
			//output the initial state of the last patch
			PrintWriter forestFile =  startPlainTextReportFile ( "forestStartFile.txt" );
			for (int i = 0; i < ecosystem.sizeX; i++) {
				int j = 0;
				String s = String.format( "%10.2f  ", loggerList.get(0).canSeeGrid.getValueAt(i, j));
				for(j = 1; j < ecosystem.sizeY; j++) {
					s += String.format( "%10.2f  ", loggerList.get(0).canSeeGrid.getValueAt(i, j));	
				}
				writeLineToReportFile( s, forestFile);
			}
			endReportFile( forestFile );
					
		}
			
		for ( NGO a : ngoList )  
			a.step();

		for ( LocalEnforcer a : localEnforcersList )  
			a.step();
		
		for ( Logger a : loggerList )  
			a.step();
		
		ministry.step();

		roadNetwork.updateMaintenance(maintenanceCostPerYearPerMinorRoadCell,maintenanceCostPerYearPerMajorRoadCell);
		roadNetwork.collectStatistics();

		
		for(int i = 0; i < numSpecies; ++i) {
			ecosystem.speciesList.get(i).updateBiomass();
			currentBiomass[i] = ecosystem.speciesList.get(i).totalBiomass;
		}
	
		if ( getTickCount() % cuttingDays == 0 && getTickCount() > 1 ) {
			
			//output the final state of the last patch
			PrintWriter forestFile =  startPlainTextReportFile ( "forestFile.txt" );
			for (int i = 0; i < ecosystem.sizeX; i++) {
				int j = 0;
				String s = String.format( "%10.2f  ", loggerList.get(0).canSeeGrid.getValueAt(i, j));
				for(j = 1; j < ecosystem.sizeY; j++) {
					s += String.format( "%10.2f  ", loggerList.get(0).canSeeGrid.getValueAt(i, j));	
				}
				writeLineToReportFile( s, forestFile);
			}
			endReportFile( forestFile );
			
			//output the final state of the last patch
			PrintWriter roadFile =  startPlainTextReportFile ( "roadFile.txt" );
			for (int i = 0; i < roadNetwork.roads.getSizeX(); i++) {
				int j = 0;
				String s = String.format( "%10.2f  ", roadNetwork.roads.getValueAt(i, j));
				for(j = 1; j < roadNetwork.roads.getSizeY(); j++) {
					s += String.format( "%10.2f  ", roadNetwork.roads.getValueAt(i, j));	
				}
				writeLineToReportFile( s, roadFile);
			}
			endReportFile( roadFile );
			
			//move to a new patch

			if ( rDebug > 0 )
				System.err.printf( "\n ******************** Moving to a new patch ****************\n" );
			
			createEcosystem();
			createRoads();
			
			for(int i = 0; i < numSpecies; ++i) {
				initialBiomass[i] = ecosystem.speciesList.get(i).totalBiomass;
				currentBiomass[i] = ecosystem.speciesList.get(i).totalBiomass;
			}
			
			for ( Logger a : loggerList ) {
				a.loggingSites = new Vector<Observation> ();
				a.canSeeGrid = new Object2DGrid(sizeX, sizeY);
				a.observedCells = new Object2DGrid(sizeX, sizeY);
				for(int i = 0; i < sizeX; i++) {
					for(int j = 0; j < sizeY; j++) {
						a.canSeeGrid.putValueAt(i, j, 0);
						a.observedCells.putValueAt(i, j, 0);
					}
				} 
				a.initialSurvey();
			}
				

			for ( LocalEnforcer a : localEnforcersList )  {
				a.monitoringSites = new Vector<Observation> ();
				a.reportSites = new Vector<Report> ();
				a.canSeeGrid = new Object2DGrid(sizeX, sizeY);
				a.observedCells = new Object2DGrid(sizeX, sizeY);
				for(int i = 0; i < sizeX; i++) {
					for(int j = 0; j < sizeY; j++) {
						a.canSeeGrid.putValueAt(i, j, 0);	
						a.observedCells.putValueAt(i, j, 0);
					}
				}
			}
			
			for ( NGO a : ngoList )  {
				a.monitoringSites = new Vector<Observation> ();
				a.reportSites = new Vector<Report> ();
				a.canSeeGrid = new Object2DGrid(sizeX, sizeY);
				a.observedCells = new Object2DGrid(sizeX, sizeY);
				for(int i = 0; i < sizeX; i++) {
					for(int j = 0; j < sizeY; j++) {
						a.canSeeGrid.putValueAt(i, j, 0);	
						a.observedCells.putValueAt(i, j, 0);
					}
				}
			}
			
			
		}
		
   		if ( rDebug > 0 )
			System.out.printf( "<== Model step done.\n" );

		stepReport();		// write aggregate measures to report file

	}


	/////////////////////////////////////////////////////////////////////////////////
	// stepReport
	// called each model time step to write out lines that look like: 
    //     timeStep  ...data...data...data...

	public void stepReport () {
		if ( rDebug > 0 )
			System.out.printf( "==> Model stepReport %.0f:\n", getTickCount() );

		calcModelStats();

		// set up a string with the values to write -- start with time step
		String s = String.format( "%5.0f  ", getTickCount() );

		// Append to String s here to write other data to report lines:

		s += String.format( " %5.2f ", (double) currentBiomass[1]/initialBiomass[1]  );
		s += String.format( " %5.2f ", (double) currentBiomass[2]/initialBiomass[2]  );
		s += String.format( " %10.2f ", (double) loggerList.get(0).loggingRevenue  );
		s += String.format( " %10.2f ", (double) loggerList.get(0).loggingCosts  );
		s += String.format( " %10.2f ", (double) loggerList.get(0).harvestFinesPaid  );
		s += String.format( " %10.2f ", (double) loggerList.get(0).bribesPaid  );
		s += String.format( " %10.2f ", (double) loggerList.get(0).bribeFinesPaid  );
		s += String.format( " %10.2f ", (double) roadNetwork.aveDistRoad  );
		s += String.format( " %10.2f ", (double) roadNetwork.percentPaved  );
		s += String.format( " %10.2f ", (double) loggerList.get(0).percRisk );
		s += String.format( " %10.2f ", (double) loggerList.get(0).m  );
		s += String.format( " %10.2f ", (double) localEnforcersList.get(0).m );
		s += String.format( " %10.2f ", (double) roadNetwork.branchPoints );
		s += String.format( " %10.2f ", (double) loggerList.get(0).expectedNetValue );
		s += String.format( " %10.2f ", (double) loggerList.get(0).laborCosts );


		// write it to the plain text report file, 'flush' buffer to file
		writeLineToPlaintextReportFile( s );
		getPlaintextReportFile().flush();

	}

	// calcModelStats
	// calculate stats and measures of interest.
	//
	// for testing:
	// - the average localEnforcers X location, store in model field.
	//
	
	public void calcModelStats () {
		
		
		/*localEnforcersAvgX = 0.0;

		for ( LocalEnforcer a : localEnforcersList )
			localEnforcersAvgX += a.getX();
		if ( localEnforcersList.size() > 1 )
			localEnforcersAvgX /= localEnforcersList.size();*/

	}

	// writeHeaderCommentsToReportFile
	// customize to match what you are writing to the report files in stepReport.
	
	public void writeHeaderCommentsToReportFile () {
		writeLineToPlaintextReportFile( "#                    " );
		writeLineToPlaintextReportFile( "#        Frac   Frac       Log         Log         Harv        Bribe       Bribe       Ave. Dist.  Percent     Logger      Logger      Monitor     Branch      Expected    Labor" );
		writeLineToPlaintextReportFile( "# time   Sp.1   Sp.2       Rev         Costs       Fines       Paid        Fines       Road        Paved       Harv.Risk   Bribe Risk  Bribe Risk  Points      Net Value   Costs" );
	}

	//////////////////////////////////////////////////////////////////////////////////
	// printProjectHelp
	// this should be filled in with some help to get from running as
	//        bin/guirun.sh -h
	//
	
	public void printProjectHelp() {
		System.out.printf( "\n%s -- CAFI ABModel1 \n", getName() );

		System.out.printf( " TBA " );
		System.out.printf( "" );



		System.out.printf( "Settable Parameterts AT START OF RUN ONLY:\n" );
		System.out.printf( "  SizeX, SizeY -- world size\n" );
		System.out.printf( "  numLocalEnforcers -- number to create in that world.\n" );
		System.out.printf( "  useTorus -- true => torus (vs bounded) world.\n" );
		System.out.printf( "  maxDensityS1 -- density of tree species, m3/ha 1.\n" );
		System.out.printf( "  maxDensityS2 -- density of tree species, m3/ha 2.\n" );
		System.out.printf( "  maxDensityS3 -- density of tree species, m3/ha 3.\n" );
		System.out.printf( "  cellSize -- size of one cell, ha.\n" );
		System.out.printf( "  maxRoadBuild -- number of road cells that can be built per timestep.\n" );
		System.out.printf( "  maintenanceCostPerYearPerKmMinorRoad -- cost to haul 1m3 of lumber, one km back to origin.\n" );
		System.out.printf( "  maintenanceCostPerYearPerKmMajorRoad -- cost to haul 1m3 of lumber, one km back to origin.\n" );
		System.out.printf( "  initLogSurveyFrac -- initial fraction of land covered by survey.\n");
		System.out.printf( "  subsLogSurveyFrac -- fraction of land covered by surveys each timestep.\n");
		System.out.printf( "  maxCost -- cost at color limits in netvalue display.\n" );
		System.out.printf( "  widthMinorRoad -- width in m when minor road is laid.\n" );
		System.out.printf( "  widthMajorRoad -- width in m when major road is laid.\n" );
		System.out.printf( "  numSpecies -- number of tree species.\n" );
		System.out.printf( "  monRandArea -- area monitor adds to list each timestep.\n" );
		System.out.printf( "  NGORandArea -- area NGO adds to list each timestep.\n" );
		System.out.printf( "  costLayMinorRoadPerKm -- cost of one minor road unit.\n" );
		System.out.printf( "  costLayMajorRoadPerKM -- cost of one major road unit.\n" );
		System.out.printf( "  costHarvestOnePeriodPerHa -- cost of one timestep spent harvesting one unit.\n" );
		System.out.printf( "  costHarvestPerM3 -- cost of harvesting 1 m3 of wood.\n" );
		System.out.printf( "  priceSpecies1PerM3 -- market price of species 1.\n" );
		System.out.printf( "  priceSpecies2PerM3 -- market price of species 2.\n" );
		System.out.printf( "  fineSpecies2PerM3 -- fine levied against cutting of species 2.\n" );
		System.out.printf( "  maxDistRoadMeters -- max distance from road that can be logged, in cells.\n" );	
		System.out.printf( "  maxLogMe3 -- max volume to log in per timestep.\n" );
		System.out.printf( "  maxMonArea -- max area to monitor/enforce per timestep.\n" );
		System.out.printf( "  maxNGOArea -- max area to monitor per timestep.\n" );
		System.out.printf( "  lengthMemory -- max number of sites to have information about.\n" );	
		System.out.printf( "  phiLogger -- per unit rate levied against loggers caught bribing.\n" );	
		System.out.printf( "  phiMonitor -- per unit rate levied against monitors caught bribing.\n" );	
		System.out.printf( "  betaBribe -- bargaining power between monitor and logger --> beta near 0 means monitor strong, vice versa.\n" );	
		System.out.printf( "  sigmaBribe -- relative standard deviation of bribes.\n" );
		System.out.printf( "  bribeOn -- whether bribes are on or not.\n" );
		System.out.printf( "  allowedCutRoads -- whether can cut trees of any kind in order to build road.\n" );
		System.out.printf( "  edgeOn -- whether there is a boundary in place to remove edge effects.\n" );
		System.out.printf( "  treeDistTestOn -- whether trees are distributed randomly (off) or in some test pattern (on).\n" );
		System.out.printf( "  riskMemory -- num timesteps agents remember bribing/fining events.\n" );
		System.out.printf( "  graphUpdateFrequency -- frequency in timesteps of graph updates.\n" );
		System.out.printf( "  maxProbCheckpointCatch -- probability of checkpoint catching fraud when truckload is all sp2.\n" );
		System.out.printf( "  useKernelSmooth -- whether we smooth out the distribution and create spatial auto-correlation.\n" );
		System.out.printf( "  usePowerDist -- whether we use power distribution or uniform distribution.\n" );
		System.out.printf( "  kernelRadius -- radius of the kernel for smoothing; weights are bell-shaped with distance.\n" );
		System.out.printf( "  kernelWeightParam -- parameter for kernel shape.\n" );
		System.out.printf( "  powerAlpha -- parameter for power distribution.\n" );
		System.out.printf( "  cuttingDays -- number of days in patch across which allowance is spaced.\n" );
		System.out.printf( "\n" );
		System.out.printf( "To compile:  bin/compile.sh\n" );
		System.out.printf( "or try this to get warnings (if any) printed out:\n" );
		System.out.printf( "   bin/compile.sh -Xlint:unchecked\n" );
		System.out.printf( "\n" );
		System.out.printf( "To run:\n" );
		System.out.printf( "   bin/guirun.sh\n" );
		System.out.printf( "   bin/guirun.sh X=50 Y=50 nLE=400\n" );
		System.out.printf( "   bin/batchrun.sh T=500 X=50 Y=50 nLE=400\n" );
		System.out.printf( "\n" );

		printParametersMap();

		System.out.printf( "\n" );

		printForDrone ();
		System.out.printf( "\n" );

		System.exit( 0 );

	}

	public void printForDrone () {


		ArrayList<String> parameterNames = new ArrayList<String>( parametersMap.values() );
		ArrayList<String> parameterAliases = new ArrayList<String>( parametersMap.keySet() );
		for( int i = 0; i < parameterAliases.size(); i++ ) {
			Method getmethod = null;
			String parAlias = (String)  parameterAliases.get(i);
			String parName = (String) parametersMap.get( parAlias );
			getmethod = findGetMethodFor( parName );
			if( getmethod != null ) {
				try {
					Object returnVal = getmethod.invoke( this, new Object[] {} );
					String s = "<value param=\"" + parName + "\" value=\"" + returnVal + "\" />";
					System.out.printf( "%s\n", s );
				} catch( Exception e ) { e.printStackTrace(); }
			}
			else {
				System.err.printf ( "COULD NOT FIND SET METHOD FOR:  %s\n", 
					parameterNames.get( i ) );
				System.err.printf ( "Is the entry in the parametersMap for this correct?" );
			}
		}

		System.out.printf( "\n" );


		for( int i = 0; i < parameterAliases.size(); i++ ) {
			Method getmethod = null;
			String parAlias = (String)  parameterAliases.get(i);
			String parName = (String) parametersMap.get( parAlias );
			getmethod = findGetMethodFor( parName );
			if( getmethod != null ) {
				try {
					Object returnVal = getmethod.invoke( this, new Object[] {} );
					String s = "<abbrev param=\"" + parName + "\" abbrev=\"" + parAlias + "\" />";
					System.out.printf( "%s\n", s );
				} catch( Exception e ) { e.printStackTrace(); }
			}
			else {
				System.err.printf ( "COULD NOT FIND SET METHOD FOR:  %s\n", 
					parameterNames.get( i ) );
				System.err.printf ( "Is the entry in the parametersMap for this correct?" );
			}
		}


		/* for regular drone:
		for( int i = 0; i < parameterAliases.size(); i++ ) {
			Method getmethod = null;
			String parAlias = (String)  parameterAliases.get(i);
			String parName = (String) parametersMap.get( parAlias );
			getmethod = findGetMethodFor( parName );
			if( getmethod != null ) {
				try {
					Object returnVal = getmethod.invoke( this, new Object[] {} );
					String s = "abbrev " + parName + " = " + parAlias ;
					System.out.printf( "%s\n", s );
				} catch( Exception e ) { e.printStackTrace(); }
			}
			else {
				System.err.printf ( "COULD NOT FIND SET METHOD FOR:  %s\n", 
					parameterNames.get( i ) );
				System.err.printf ( "Is the entry in the parametersMap for this correct?" );
			}
		}
		*/

	}

	///////////////////////////////////////////////////////////////////////////////
	// moveObjectInWorld
	// try to move object to the requested dx,dy .
	// note we must check to be sure the new location is "legal",
	// i.e., cell is empty and cell is not off the edge of the world.
	// NB: if either new values is illegal, don't move.
	// If move is ok:
	// - tell the world the object is not where it was
	// - tell the world the object is now at the new location
	// Return: null if not moved, otherwise
	//         return a Point object with new x,y values.
	// NOTE: its up to the caller to tell the object it moved.

	public Point moveObjectInWorld ( ObjectInGrid obj, int dX, int dY ) {

		int newX = obj.getX() + dX;	// get location it wants to move to
		int newY = obj.getY() + dY;

		// first check to be sure new location is in the world!
		//if ( newX < 0 || newY < 0 || newX >= sizeX || newY >= sizeY ) {
		//	if ( rDebug > 0 )
		//		System.out.printf( "   - obj tried to move offworld (%d,%d).\n", 
		//					   newX, newY );
		//	return null;
		//}
		newX = (sizeX + newX) % sizeX;
		newY = (sizeY + newY) % sizeY;

		// see if new cell is empty.
		if ( world.getObjectAt( newX, newY ) != null ) {
			if ( rDebug > 0 )
				System.out.printf( "   - tried to move to occupied cell (%d,%d)\n", 
							   newX, newY );
			return null;
		}

		// its ok to move, so tell the world and the object about the move
		world.putObjectAt( obj.getX(), obj.getY(), null );	// old cell empty now
		world.putObjectAt( newX, newY, obj );			   	// obj in new cell now
		Point loc = new Point( newX, newY );   // store new location values in Point

		return loc;
	}


	/////////////////////////////////////////////////////////////////////////////
	// setters and getters 
	// These must be defined to be able to set parameters via GUI and run command 
	//
	// NB: some things can't be changed after run starts (eg numAnts, sizeX,Y etc)

	public int getNumLocalEnforcers () { return numLocalEnforcers; }
	public void setNumLocalEnforcers ( int n ) { numLocalEnforcers = n; }
	public int getSizeX () { return sizeX; }
	public void setSizeX ( int szX ) { sizeX = szX; }
	public int getSizeY () { return sizeY; }
	public void setSizeY ( int szY ) { sizeY = szY;  }
	public int getMaxDistRoad () { return maxDistRoad; }
	public void setMaxDistRoad ( int mDR ) { maxDistRoad = mDR;  }
	public int getMaxMonCells () { return maxMonCells; }
	public void setMaxMonCells ( int mMC ) { maxMonCells = mMC;  }
	public int getMaxNGOCells () { return maxNGOCells; }
	public void setMaxNGOCells ( int mMC ) { maxNGOCells = mMC;  }
	public int getLengthMemory () { return lengthMemory; }
	public void setLengthMemory ( int lM ) { lengthMemory = lM;  }
	public double getSubsLogSurveyFrac () { return subsLogSurveyFrac; }
	public void setSubsLogSurveyFrac ( double sF ) { subsLogSurveyFrac = sF;  }
	public double getInitLogSurveyFrac () { return initLogSurveyFrac; }
	public void setInitLogSurveyFrac ( double sF ) { initLogSurveyFrac = sF;  }
	public double getMaintenanceCostPerYearPerMinorRoadCell () { return maintenanceCostPerYearPerMinorRoadCell; }
	public void setMaintenanceCostPerYearPerMinorRoadCell ( double tC ) { maintenanceCostPerYearPerMinorRoadCell = tC;  }
	public double getMaintenanceCostPerYearPerMajorRoadCell () { return maintenanceCostPerYearPerMajorRoadCell; }
	public void setMaintenanceCostPerYearPerMajorRoadCell ( double tC ) { maintenanceCostPerYearPerMajorRoadCell = tC;  }
	public double getMaxVolumeS1 () { return maxVolumeS1; }
	public void setMaxVolumeS1 ( double maxD ) { maxVolumeS1 = maxD;  }
	public double getPowerAlpha () { return powerAlpha; }
	public void setPowerAlpha ( double k ) { powerAlpha = k;  }
	public double getMaxVolumeS2 () { return maxVolumeS2; }
	public void setMaxVolumeS2 ( double maxD ) { maxVolumeS2 = maxD;  }
	public double getMaxVolumeS0 () { return maxVolumeS0; }
	public void setMaxVolumeS0 ( double maxD ) { maxVolumeS0 = maxD;  }
	public double getFracCutMinorRoad () { return fracCutMinorRoad; }
	public void setFracCutMinorRoad ( double fC ) { fracCutMinorRoad = fC;  }
	public double getFracCutMajorRoad () { return fracCutMajorRoad; }
	public void setFracCutMajorRoad ( double fC ) { fracCutMajorRoad = fC;  }
	public double getMaxProbCheckpointCatch () { return maxProbCheckpointCatch; }
	public void setMaxProbCheckpointCatch ( double mPCC ) { maxProbCheckpointCatch = mPCC;  }
	public double getMaxCost () { return maxCost; }
	public void setMaxCost ( double maxC ) { maxCost = maxC;  }
	public int getMonRandCells () { return monRandCells; }
	public void setMonRandCells ( int monRC ) { monRandCells = monRC;  }
	public int getNGORandCells () { return NGORandCells; }
	public void setNGORandCells ( int monRC ) { NGORandCells = monRC;  }
	public int getNumSpecies () { return numSpecies; }
	public void setNumSpecies ( int numS ) { numSpecies = numS;  }
	public boolean getUseTorus () { return useTorus; }
	public void setUseTorus ( boolean b ) { useTorus = b; }
	public boolean getAllowedCutRoads () { return allowedCutRoads; }
	public void setAllowedCutRoads ( boolean b ) { allowedCutRoads = b; }
	public boolean getUseKernelSmooth () { return useKernelSmooth; }
	public void setUseKernelSmooth ( boolean b ) { useKernelSmooth = b; }
	public boolean getUsePowerDist () { return usePowerDist; }
	public void setUsePowerDist ( boolean b ) { usePowerDist = b; }
	public int getCuttingDays () { return cuttingDays; }
	public void setCuttingDays ( int b ) { cuttingDays = b; }
	public double getCostHarvestPerM3 () { return costHarvestPerM3; }
	public void setCostHarvestPerM3 ( double b ) { costHarvestPerM3 = b; }

    public double getMaintenanceCostPerYearPerKmMinorRoad () { return maintenanceCostPerYearPerKmMinorRoad;}
    public double getMaintenanceCostPerYearPerKmMajorRoad () { return maintenanceCostPerYearPerKmMajorRoad;}
    public double getWidthMinorRoad () { return widthMinorRoad;}
    public double getWidthMajorRoad () { return widthMajorRoad;}
	public double getMaxDensityS1 () { return maxDensityS1;}
	public double getMaxDensityS2 () { return maxDensityS2;}
	public double getMaxDensityS0 () { return maxDensityS0;}
	public double getMonRandArea () { return monRandArea;}
	public double getNGORandArea () { return monRandArea;}
	public double getCostLayMinorRoadPerKm () { return costLayMinorRoadPerKm;}
	public double getCostLayMajorRoadPerKm () { return costLayMajorRoadPerKm;}
	public double getMaxDistRoadMeters () { return maxDistRoadMeters;}
	public double getMaxLogM3 () { return maxLogM3;}
	public double getMaxMonArea () { return maxMonArea;}
	public double getMaxNGOArea () { return maxNGOArea;}
	public double getCellSize () { return cellSize;}
	public double getKernelWeightParam () { return kernelWeightParam;}

    public void setMaintenanceCostPerYearPerKmMinorRoad (double a) { maintenanceCostPerYearPerKmMinorRoad = a; }
    public void setMaintenanceCostPerYearPerKmMajorRoad (double a) { maintenanceCostPerYearPerKmMajorRoad = a; }
    public void setWidthMinorRoad (double a) { widthMinorRoad = a; }
    public void setWidthMajorRoad (double a) { widthMajorRoad = a; }
	public void setMaxDensityS1 (double a) { maxDensityS1 = a; }
	public void setMaxDensityS2 (double a) { maxDensityS2 = a; }
	public void setMaxDensityS0 (double a) { maxDensityS0 = a; }
	public void setMonRandArea (double a) { monRandArea = a; }
	public void setNGORandArea (double a) { NGORandArea = a; }
	public void setCostLayMinorRoadPerKm (double a) { costLayMinorRoadPerKm = a; }
	public void setCostLayMajorRoadPerKm (double a) { costLayMajorRoadPerKm = a; }
	public void setMaxDistRoadMeters (double a) { maxDistRoadMeters = a; }
	public void setMaxLogM3 (double a) { maxLogM3 = a; }
	public void setMaxMonArea (double a) { maxMonArea = a; }
	public void setMaxNGOArea (double a) { maxNGOArea = a; }
	public void setCellSize (double a) { cellSize = a; }
	public void setKernelWeightParam (double a) { kernelWeightParam = a; }
	
	public double getCostLayMinorRoad() { return costLayMinorRoad; }
	public void setCostLayMinorRoad(double x) {costLayMinorRoad = x;}
	public double getCostLayMajorRoad() { return costLayMajorRoad; }
	public void setCostLayMajorRoad(double x) { costLayMajorRoad = x;}
	public double getMaxHarvestOnePeriod() { return maxHarvestOnePeriod; }
	public void setMaxHarvestOnePeriod(double x) {maxHarvestOnePeriod = x;}
	public double getPriceSpecies1PerM3() { return priceSpecies1PerM3; }
	public void setPriceSpecies1PerM3(double x) { priceSpecies1PerM3 = x;}
	public double getPriceSpecies2PerM3() { return priceSpecies2PerM3; }
	public void setPriceSpecies2PerM3(double x) { priceSpecies2PerM3 = x;}
	public double getFineSpecies2PerM3() { return fineSpecies2PerM3; }
	public void setFineSpecies2PerM3(double x) { fineSpecies2PerM3 = x;}
	public double getPhiLogger() { return phiLogger; }
	public void setPhiLogger(double x) { phiLogger = x;}
	public double getPhiMonitor() { return phiMonitor; }
	public void setPhiMonitor(double x) { phiMonitor = x;}
	public double getBetaBribe() { return betaBribe; }
	public void setBetaBribe(double x) { betaBribe = x;}
	public double getSigmaBribe() { return sigmaBribe; }
	public void setSigmaBribe(double x) { sigmaBribe = x;}
	public boolean getBribeOn() { return bribeOn; }
	public void setBribeOn(boolean x) { bribeOn = x;}
	public boolean getEdgeOn() { return edgeOn; }
	public void setEdgeOn(boolean x) { edgeOn = x;}
	public int getRiskMemory() { return riskMemory; }
	public boolean getTreeDistTestOn() { return treeDistTestOn; }
	public void setTreeDistTestOn(boolean x) { treeDistTestOn = x;}
	public void setRiskMemory(int x) { riskMemory = x;}
	public int getGraphUpdateFrequency() { return graphUpdateFrequency; }
	public void setGraphUpdateFrequency(int x) { graphUpdateFrequency = x;}
	public int getKernelRadius() { return kernelRadius; }
	public void setKernelRadius(int x) { kernelRadius = x;}
	
	// getters for aggregate measures
	public double getLocalEnforcersAvgX() { return localEnforcersAvgX; }


	///////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////
	//
	//            USUALLY NO NEED TO CHANGE THINGS BELOW HERE
	//
	///////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////



	////////////////////////////////////////////////////////////////////////////
	// constructor, if need to do anything special.
	public Model () {
	}

	///////////////////////////////////////////////////////////////////////////
	// setup
	// set defaults after a run start or restart

	public void setup () {
		schedule = null;
		if ( rDebug > 1 )		// won't print at startup: cmdline pars done below  
			System.out.printf( "==> Model-setup...\n" );

		userSetup();

		System.gc ();   // garabage collection of discarded objects
		super.setup();  // THIS SHOULD BE CALLED after setting defaults in setup().
						// This processes cmd line parameter settings

		schedule = new Schedule (1);  // create AFTER calling super.setup()

		if ( rDebug > 1 )
			System.out.printf( "\n<=== Model-setup() done.\n" );

	}

	///////////////////////////////////////////////////////////////////////////
	// buildModel
	// We build the "conceptual" parts of the model.
	// (vs the display parts, and the schedule)
	//
	// Create a 2D world, tell the Ants about it.
	// Create ants, put them in the world, one per cell,
	// and also add them to the antList.

	public void buildModel () {
		if ( rDebug > 1 )
			System.out.printf( "==> buildModel...\n" );

		// CALL FIRST -- defined in super class -- it starts RNG, etc
		buildModelStart();

		userBuildModel();

		// some post-load finishing touches
		startReportFile();

		// you probably don't want to remove any of the following
		// calls to process parameter changes and write the
		// initial state to the report file.
		// NB -> you might remove/add more agentChange processing
        applyAnyStoredChanges();
        stepReport();
        getReportFile().flush();
        getPlaintextReportFile().flush();

		if ( rDebug > 1 )
			System.out.printf( "<== buildModel done.\n" );
	}




	//////////////////////////////////////////////////////////////////////////////
	public Schedule getSchedule () {	return schedule; }

	public String getName () { return "Model"; }



	/////////////////////////////////////////////////////////////////////////////
	// processEndOfRun
	// called once, at end of run.
	// writes some final info, closes report files, etc.
	public void processEndOfRun ( ) {
		if ( rDebug > 0 )  
			System.out.printf("\n\n===== processEndOfRun =====\n\n" );
		applyAnyStoredChanges();
		endReportFile();

		this.fireStopSim();
	}

}
