package CAFI;

/**
NGO.java 

Class for NGO, a type of Agent

**/

import java.util.Collections;
import java.util.Comparator;
import java.util.Formatter;
import java.util.Vector;
import java.awt.Point;
import java.util.ArrayList;

import uchicago.src.sim.util.*;
import uchicago.src.sim.gui.*;
import uchicago.src.sim.space.*;
import uchicago.src.sim.gui.ColorMap;
import java.awt.Color;


public class NGO extends Agent {
	// "class" variables -- one value for all instances 
    public  static int          	nextId = 0; // to give each a unique id

	// fields (instance variables)  
	int 				id;		// unique id number for each instance


	public Vector<Observation> monitoringSites;
	public Vector<Report> reportSites;
	public Object2DGrid canSeeGrid;
	public Object2DGrid observedCells;
	public double		finesCollected;
	public double		observedIllegalCutting;

	// a constructor
	public NGO ( ) {
		id = nextId++;
		
		finesCollected = 0;
		canSeeGrid = new Object2DGrid (model.sizeX, model.sizeY);
		monitoringSites = new Vector<Observation> ();
		reportSites = new Vector<Report> ();
		canSeeGrid = new Object2DGrid(model.sizeX, model.sizeY);
		observedCells = new Object2DGrid(model.sizeX, model.sizeY);
		for(int i = 0; i < model.sizeX; i++) {
			for(int j = 0; j < model.sizeY; j++) {
				canSeeGrid.putValueAt(i, j, 0);	
				observedCells.putValueAt(i, j, 0);
			}
		}
		observedIllegalCutting = 0;
	}

	////////////////////////////////////////////////////////////////////////////
	// step
	// Top level definition of what the agent can do each time its activated.

	public void step () {
		if ( model.getRDebug() > 0 ) 
			System.err.printf( "   -->NGO-step() for id=%d.\n", id );

		lookForest ();

		trimMonitoringSitesMemory ();

		if ( model.getRDebug() > 1 ) 
			System.err.printf("    <--NGO-step() done.\n" );
	}
	
	////////////////////////////////////////////////////////////
	//look at the forest to see where we should monitor
	public void lookForest() {


		//update and enforce the most populous cells
		int limit = 0;
		if (model.maxNGOCells > monitoringSites.size()) {
			limit = monitoringSites.size();
		}else {
			limit = model.maxNGOCells;
		}
		for(int i = 0; i < limit; ++i) {
			
			double oldObsSpecies2 = (double) monitoringSites.get(i).massSpecies2;
			Observation currentObs = monitoringSites.get(i);
			currentObs.updateObservation();
			canSeeGrid.putValueAt(currentObs.x, currentObs.y, currentObs.massSpecies2);
			double currentObsSpecies2 = (double) monitoringSites.get(i).massSpecies2;
			
			if (currentObsSpecies2 < oldObsSpecies2) { //trees have been illegally cut!!
				
				observedIllegalCutting += (oldObsSpecies2 - currentObsSpecies2);
				Report newReport = new Report(currentObs.x, currentObs.y, oldObsSpecies2 - currentObsSpecies2);
				reportSites.add(newReport);
			} else {
				Report newReport = new Report(currentObs.x, currentObs.y, 0);
				reportSites.add(newReport);
			}

			
		}
		
		//number of new cells to survey
		int numSeeCells =  model.NGORandCells;
		
		//randomly select a subset of cells to look at
		for(int i = 0; i < numSeeCells; ++i) {

			int newX = ModelParameters.getUniformIntFromTo(0, model.sizeX - 1);
			int newY = ModelParameters.getUniformIntFromTo(0, model.sizeY - 1);
			if(observedCells.getValueAt(newX, newY) == 0) {
				//if they are cells we haven't seen yet, add them to the list
				Observation newObs = new Observation(newX,newY);
				monitoringSites.add(newObs);
				observedCells.putValueAt(newX, newY, 1);
				canSeeGrid.putValueAt(newX, newY, newObs.massSpecies2);
			}
		}
		
		//order them from highest to lowest mass of species 2
		removeDuplicates();
	}


	//sort, and remove duplicates from the vector of memory
	public void removeDuplicates () {
		int i = 0;
		Collections.sort( monitoringSites, (java.util.Comparator<? super Observation>) new ObsSpecies2Comparator() );
		while (i < monitoringSites.size()-1) {
			if(monitoringSites.get(i).x == monitoringSites.get(i+1).x && monitoringSites.get(i).y == monitoringSites.get(i+1).y) {
				monitoringSites.removeElementAt(i+1);
			} else {
				i = i + 1;
			}
		}
	}
	
	//trim down the vector of memory
	public void trimMonitoringSitesMemory () {
		while (monitoringSites.size() > model.lengthMemory) {
			int currentSize = monitoringSites.size()-1;
			int currentX = monitoringSites.get(currentSize).x;
			int currentY = monitoringSites.get(currentSize).y;
			observedCells.putValueAt(currentX, currentY, 0);
			canSeeGrid.putValueAt(currentX, currentY, 0);
			monitoringSites.removeElementAt(currentSize);
			
		}
	}
	
	private class ObsSpecies2Comparator implements Comparator {
	       public final int compare ( Object obs1, Object obs2 ) {
	           double d1 = ( (Observation) obs1 ).massSpecies2;
	           double d2 = ( (Observation) obs2 ).massSpecies2;
	           if ( d1 < d2 ) //orders values descending...
	               return 1;
	           else if ( d1 > d2 )
	               return -1;
	           return 0;
	       }
	   }
	
	////////////////////////////////////////////////////////////
	// printOut
	// print fields to System.out.
	//
	// Note: we can use the class's field variables in all methods.

	public void printOut ( ) {
		System.out.printf( " - LocalEnforcer id %2d (x,y=%d,%d; live=%b).\n",
						   id );
	}


}
