package CAFI;

/**
Observation.java 

A class that can be root class for observations of all kinds.

**/

import java.util.Formatter;
import java.awt.Point;
import java.util.ArrayList;

import uchicago.src.sim.util.*;
import uchicago.src.sim.gui.*;

public class Observation {
	// "class" variables -- one value for all instances 
    public  static int          	nextId = 0; // to give each a unique id
	public  static Model		   	model;      // the model "in charge"
	public  static GridWorld	   	world;  	// where (some) agents live

	// fields (instance variables)  
	int 				id;		// unique id number for each instance
	int					x;
	int					y;
	double				massSpecies0;
	double				massSpecies1;
	double				massSpecies2;
	double 				netValue;
	double				netValueSp1Only;
	double				netValueSp1Sp2;
	double				netReturnSp1M3;
	double				netReturnSp1Sp2M3;
	double 				maxNetReturnM3;
	boolean				flag0Sp1Only1Both;
	double				probSelection;
	int					elementID;
	
	


	// a constructor 
	
	//for now, let the observation include all observable variables
	public Observation (int newX, int newY ) {
		id = nextId++;
		x = newX;
		y = newY;
		massSpecies0 = model.ecosystem.getTree(0).getValueAt(x, y);
		massSpecies1 = model.ecosystem.getTree(1).getValueAt(x, y);
		massSpecies2 = model.ecosystem.getTree(2).getValueAt(x, y);
		elementID = (y)*model.sizeX + x;
		

		
		netValue = calcNetValue();

	}

	public double calcNetValue () { 
		
		double deltaRSp1 = 0;
		double deltaRSp2 = 0;
		if(model.loggerList.get(0).currentLoggedSp1 + model.loggerList.get(0).currentLoggedSp2 > 0) {
			deltaRSp1 = massSpecies1/(model.loggerList.get(0).currentLoggedSp1 + model.loggerList.get(0).currentLoggedSp2);
			deltaRSp2 = massSpecies2/(model.loggerList.get(0).currentLoggedSp1 + model.loggerList.get(0).currentLoggedSp2 + massSpecies1);
		}
		
		double changeCostsSp1 = deltaRSp1*model.maxProbCheckpointCatch*(model.loggerList.get(0).currentLoggedSp2)*model.fineSpecies2PerM3;
		double changeCostsSp2 = -deltaRSp2*model.maxProbCheckpointCatch*(model.loggerList.get(0).currentLoggedSp2 + massSpecies2)*model.fineSpecies2PerM3;
		
    	double valueSpecies2  = (model.priceSpecies2PerM3- model.fineSpecies2PerM3 * model.loggerList.get(0).percRisk - model.costHarvestPerM3) * massSpecies2;// + changeCostsSp2;
    	//if (valueSpecies2 < 0) {
    	//	valueSpecies2 = 0;
   
    	//}
		/*int numDaysLabor = ((int) ((massSpecies1 + massSpecies2) / model.maxHarvestOnePeriod))+1;
		if (valueSpecies2 < 0) {
			numDaysLabor = ((int) ((massSpecies1) / model.maxHarvestOnePeriod))+1;
		}*/
		int numDaysLabor = 1;
		double costRoad = model.costLayMinorRoad;
	
		
		int nearestRoadX = model.roadNetwork.roadListX.get((int) model.roadNetwork.closestRoadNode.getValueAt(x, y));
		int nearestRoadY = model.roadNetwork.roadListY.get((int) model.roadNetwork.closestRoadNode.getValueAt(x, y));
		double costMaint = model.roadNetwork.cumMaintenanceCostToAccess.getValueAt(nearestRoadX,nearestRoadY);
		
		//double currentNetValue =   //- (Math.max(model.roadNetwork.distanceToRoads.getValueAt(x, y) - model.maxDistRoad,0)) * model.costLayMinorRoad  + 
		//							- model.costHarvestOnePeriod * numDaysLabor +
	//								model.priceSpecies1PerM3 * massSpecies1 + changeCostsSp1
		//							- costMaint +
			//						(valueSpecies2);
		
		
		netValueSp1Sp2 =  	//- (Math.max(model.roadNetwork.distanceToRoads.getValueAt(x, y) - model.maxDistRoad,0)) * model.costLayMinorRoad  + 
							- model.costHarvestPerM3 * massSpecies1 +
							model.priceSpecies1PerM3 * massSpecies1 //;+ changeCostsSp1
							- costMaint +
							+ (valueSpecies2);
		
		netValueSp1Only =  //	- (Math.max(model.roadNetwork.distanceToRoads.getValueAt(x, y) - model.maxDistRoad,0)) * model.costLayMinorRoad  + 
									- model.costHarvestPerM3 * massSpecies1 +
									model.priceSpecies1PerM3 * massSpecies1 //;+ changeCostsSp1;
									- costMaint;
		
		double currentNetValue = Math.max(netValueSp1Sp2, netValueSp1Only);
		if(massSpecies1 > 0) {
			netReturnSp1M3 = netValueSp1Only / massSpecies1;
		} else {
			netReturnSp1M3 = 0;
		}
		if(massSpecies1 + massSpecies2 > 0) {
			netReturnSp1Sp2M3 = netValueSp1Sp2 / (massSpecies1 + massSpecies2);
		} else {
			netReturnSp1Sp2M3 = 0;
		}
			
		maxNetReturnM3 = Math.max(netReturnSp1M3, netReturnSp1Sp2M3);
		if(netReturnSp1M3 > netReturnSp1Sp2M3) {
			flag0Sp1Only1Both = false;
		} else {
			flag0Sp1Only1Both = true;
		}
		
		return currentNetValue;
	}
	
	public void updateObservation () {
		massSpecies0 = model.ecosystem.getTree(0).getValueAt(x, y);
		massSpecies1 = model.ecosystem.getTree(1).getValueAt(x, y);
		massSpecies2 = model.ecosystem.getTree(2).getValueAt(x, y);
		netValue = calcNetValue();
	}

	public void updateObservationValue () { //this is to update the netValue for the logger survey (which might change due to road placement)
											//without explicitly giving new information about the tree volume.  this should probably be changed in future
											//to avoid the use of netvaluesurface entirely.
		netValue = calcNetValue();
	}	
	////////////////////////////////////////////////////////////
	// printOut
	// print fields to System.out.
	//
	// Note: we can use the class's field variables in all methods.

	public void printOut ( ) {
		System.out.printf( " - Agent id %2d).\n",
						   id );
	}

	////////////////////////////////////////////////////////////////////////////
	// setters and getters
	//
	public int getId() {  return id; }

	public int getX() { return x; }
	public void setX( int l ) { x = l; }
	public int getY() { return y; }
	public void setY( int l ) { y = l; }

	// note these are class (static) methods, to set class (static) variables
	public static void setWorld( GridWorld w ) {	world = w; }
	public static void setModel( Model m ) { model = m; }
	public static void setNextId ( int i ) { nextId = i; }

}
