/**
 * RoadNetwork.java 
 */

package CAFI;

import uchicago.src.sim.space.Discrete2DSpace;
import uchicago.src.sim.space.Object2DGrid;

import java.util.Collections;
import java.util.Comparator;
import java.util.Vector;


public class RoadNetwork {
	// class variables
    public  static int          	nextID = 0; // to give each an ID
    public  static Model	model;
    public  static GridWorld world;

    
	// instance variables  
	int 				ID;
	public  Object2DGrid roads;
	public  Object2DGrid distanceToRoads; 
	public  Object2DGrid distanceAlongRoads;
	public  Object2DGrid roadNetworkNeighborsOut;
	public  Object2DGrid roadNetworkNeighborsIn;
	public  Object2DGrid cumMaintenanceCostToAccess;
	public 	Object2DGrid maintenanceCost;
	public  Object2DGrid closestRoadNode;
	Vector<Integer>			roadListX;
	Vector<Integer>			roadListY;
	Vector<Integer>			minorRoadListX;
	Vector<Integer>			minorRoadListY;
	Vector<Vector<Integer>> adjListRoad;
	public double			percentPaved;
	public double			aveDistRoad;	
	public int				branchPoints;
	public int				originX;
	public int 				originY;
	

	// the Road constructor
	public RoadNetwork () {
		ID = nextID++;
		
		roadListX = new Vector<Integer>();
		roadListY = new Vector<Integer>();
		minorRoadListX = new Vector<Integer>();
		minorRoadListY = new Vector<Integer>();
		adjListRoad = new Vector<Vector<Integer>> ();
		roads = new Object2DGrid ( model.sizeX, model.sizeY);

		distanceToRoads = new Object2DGrid ( model.sizeX, model.sizeY);
		distanceAlongRoads = new Object2DGrid ( model.sizeX, model.sizeY);
		cumMaintenanceCostToAccess = new Object2DGrid ( model.sizeX, model.sizeY);
		maintenanceCost = new Object2DGrid ( model.sizeX, model.sizeY);
		roadNetworkNeighborsOut = new Object2DGrid ( model.sizeX, model.sizeY);
		roadNetworkNeighborsIn = new Object2DGrid ( model.sizeX, model.sizeY);
		closestRoadNode = new Object2DGrid ( model.sizeX, model.sizeY);
		for(int i = 0; i < model.sizeX; i++) {
			for(int j = 0; j < model.sizeY; j++) {
				roads.putValueAt(i, j, 0);	
				distanceToRoads.putValueAt(i, j, -9999);
				distanceAlongRoads.putValueAt(i, j, 0);	
				cumMaintenanceCostToAccess.putValueAt(i, j, 0);	
				closestRoadNode.putValueAt(i, j, -9999);	
				maintenanceCost.putValueAt(i, j, 0);	
				Integer[] neighborListOut = new Integer[] {0,0,0,0};
				Integer[] neighborListIn = new Integer[] {0,0,0,0};
				roadNetworkNeighborsOut.putObjectAt(i, j, neighborListOut);
				roadNetworkNeighborsIn.putObjectAt(i, j, neighborListIn);
			}
		}
	}

	//Methods
	
	//maintainRoad(x, y)
	public void maintainRoad(int x, int y) {
		Vector<Integer> maintenanceListX = new Vector<Integer> ();
		Vector<Integer> maintenanceListY = new Vector<Integer> ();
		
		maintenanceListX.add(x);
		maintenanceListY.add(y);
		
		//double totalMaintenanceCost = cumMaintenanceCostToAccess.getValueAt(x, y);
		while(maintenanceListX.size() > 0) {
			int currentX = maintenanceListX.get(0);
			int currentY = maintenanceListY.get(0);
			cumMaintenanceCostToAccess.putValueAt(currentX, currentY, 0);
			maintenanceCost.putValueAt(currentX, currentY, 0);
			
			Integer[] neighborList = (Integer[]) roadNetworkNeighborsIn.getObjectAt(currentX,currentY);
			
			if(neighborList[0] == 1) {
				maintenanceListX.add(currentX-1);
				maintenanceListY.add(currentY);
			}
			if(neighborList[1] == 1) {
				maintenanceListX.add(currentX);
				maintenanceListY.add(currentY+1);
			}
			if(neighborList[2] == 1) {
				maintenanceListX.add(currentX+1);
				maintenanceListY.add(currentY);
			}
			if(neighborList[3] == 1) {
				maintenanceListX.add(currentX);
				maintenanceListY.add(currentY-1);
			}
			
			maintenanceListX.remove(0);
			maintenanceListY.remove(0);
		}
	}
	
	//updateMaintenance()
	public void updateMaintenance(double maintMin, double maintMaj) {
		Vector<Integer> maintenanceListX = new Vector<Integer> ();
		Vector<Integer> maintenanceListY = new Vector<Integer> ();
		
		maintenanceListX.add(originX);
		maintenanceListY.add(originY);
		double sourceMaintenance = 0;
		double currentMaintenance = 0;
		while(maintenanceListX.size() > 0) {
			int currentX = maintenanceListX.get(0);
			int currentY = maintenanceListY.get(0);
			Integer[] neighborList = (Integer[]) roadNetworkNeighborsOut.getObjectAt(currentX,currentY);
			
			currentMaintenance = 0;
			sourceMaintenance = 0;
			currentMaintenance += maintenanceCost.getValueAt(currentX, currentY);
			sourceMaintenance = cumMaintenanceCostToAccess.getValueAt(currentX, currentY);
			int currentRoadType = (int) roads.getValueAt(currentX, currentY);
			if( currentRoadType == 1) {
				currentMaintenance += maintMin;
			} else if (currentRoadType == 2) {
				currentMaintenance += maintMaj;
			}
			
			maintenanceCost.putValueAt(currentX, currentY, currentMaintenance);
			sourceMaintenance += currentMaintenance;
			sourceMaintenance = Math.min(sourceMaintenance, model.costLayMinorRoad); //shouldn't cost more than rebuilding the road!!!
			
			if(neighborList[0] == 1) {
				maintenanceListX.add(currentX-1);
				maintenanceListY.add(currentY);
				cumMaintenanceCostToAccess.putValueAt(currentX-1, currentY, sourceMaintenance);
			}
			if(neighborList[1] == 1) {
				maintenanceListX.add(currentX);
				maintenanceListY.add(currentY+1);
				cumMaintenanceCostToAccess.putValueAt(currentX, currentY+1, sourceMaintenance);
			}
			if(neighborList[2] == 1) {
				maintenanceListX.add(currentX+1);
				maintenanceListY.add(currentY);
				cumMaintenanceCostToAccess.putValueAt(currentX+1, currentY, sourceMaintenance);
			}
			if(neighborList[3] == 1) {
				maintenanceListX.add(currentX);
				maintenanceListY.add(currentY-1);
				cumMaintenanceCostToAccess.putValueAt(currentX, currentY-1, sourceMaintenance);
			}
			
			maintenanceListX.remove(0);
			maintenanceListY.remove(0);
		}
	}
	//buildRoad()
    public void buildRoad (int x, int y, int type, int source) {
    	
    	
    	roads.putValueAt(x, y, type);
		roadListX.add(x);
		roadListY.add(y);   //for evaluating road nodes as starting points for new roads, we use both major and minor
							//however, for evaluating DISTANCE to roads, we only use minor roads.  ie, major roads are NOT
							//a point of access to the forest.
		int currentNode = roadListX.size()-1;
		closestRoadNode.putValueAt(x, y, currentNode);
		
		//update the point from which this new road came
		if(source == 0) {
			Integer[] neighborList = (Integer[]) roadNetworkNeighborsOut.getObjectAt(x-1, y); //source is to the left
			neighborList[2] = 1; //source has an outlet to the right
			roadNetworkNeighborsOut.putObjectAt(x-1, y, neighborList);
			
			neighborList = (Integer[]) roadNetworkNeighborsIn.getObjectAt(x, y); //source is to the left
			neighborList[0] = 1; //point has an inlet from the left
			roadNetworkNeighborsIn.putObjectAt(x, y, neighborList);
			
		} else if (source == 1) {
			Integer[] neighborList = (Integer[]) roadNetworkNeighborsOut.getObjectAt(x, y+1); //neighbor is above
			neighborList[3] = 1; //source has an outlet below
			roadNetworkNeighborsOut.putObjectAt(x, y+1, neighborList);

			neighborList = (Integer[]) roadNetworkNeighborsIn.getObjectAt(x, y); //source is to the left
			neighborList[1] = 1; //point has an inlet from above
			roadNetworkNeighborsIn.putObjectAt(x, y, neighborList);
		
		} else if (source == 2) {
			Integer[] neighborList = (Integer[]) roadNetworkNeighborsOut.getObjectAt(x+1, y); //neighbor is to the right
			neighborList[0] = 1; //source has an outlet to the left
			roadNetworkNeighborsOut.putObjectAt(x+1, y, neighborList);
			
			neighborList = (Integer[]) roadNetworkNeighborsIn.getObjectAt(x, y); //source is to the left
			neighborList[2] = 1; //point has an inlet from the right
			roadNetworkNeighborsIn.putObjectAt(x, y, neighborList);
		} else if (source == 3) {
			Integer[] neighborList = (Integer[]) roadNetworkNeighborsOut.getObjectAt(x, y-1); //neighbor is below
			neighborList[1] = 1; //source has an outlet above
			roadNetworkNeighborsOut.putObjectAt(x, y-1, neighborList);
			
			neighborList = (Integer[]) roadNetworkNeighborsIn.getObjectAt(x, y); //source is to the left
			neighborList[3] = 1; //point has an inlet from below
			roadNetworkNeighborsIn.putObjectAt(x, y, neighborList);
		}
		
    	switch (type) {
    	case 5: // 'Virtual Road'
				// Don't cut anything.  This is just a trick when borders are being used to eliminate edge effects
			    // Virtual roads make the benefits of turning left or right at an edge the same or less as starting off 
				// fresh from the main road.
				break; 
		case 1: for(int i = 0; i < model.numSpecies; ++i) { //Minor Road
    				double volume = model.ecosystem.getTree(i).getValueAt(x, y);
	    			model.ecosystem.getTree(i).putValueAt(x, y, (double) volume * (1 - model.fracCutMinorRoad));
	    			minorRoadListX.add(x);
	    			minorRoadListY.add(y);
	    		}
    			break;
    	case 2: for(int i = 0; i < model.numSpecies; ++i) { //Major Road
    				double volume = model.ecosystem.getTree(i).getValueAt(x, y);
					model.ecosystem.getTree(i).putValueAt(x, y, (double) volume * (1 - model.fracCutMajorRoad));
				}
    			break;
    	}
    

    }
    
    //updateDistance(int x, int y) for updates
    public void updateDistance(int x, int y) {
		int[] neighborsX = new int[] {0,0,1,0,-1};
		int[] neighborsY = new int[] {0,1,0,-1,0};
		//int[] neighborsX = new int[] {0,0,1,0,-1,-1,1,1,-1};
		//int[] neighborsY = new int[] {0,1,0,-1,0,-1,1,-1,1};

		if(distanceToRoads.getValueAt(x,y) != 1) {
			int errFlag = 1; //this shouldn't happen since roads should always be built adjacent to existing roads
		}
		
		double currentRoadNode = closestRoadNode.getValueAt(x, y);
		
		Vector<Integer> tempX = new Vector<Integer> (0);
		Vector<Integer> tempY = new Vector<Integer> (0);
		tempX.add(x);
		tempY.add(y);
		double currentDistance = 0;
		distanceToRoads.putValueAt(tempX.get(0), tempY.get(0), currentDistance);
		
		
		while(tempX.size() > 0) {
			currentDistance = distanceToRoads.getValueAt(tempX.get(0), tempY.get(0));

			for(int j = 0; j < neighborsX.length; ++j) { //for each of the moore neighbors of the road cell
				int currentX = tempX.get(0)+neighborsX[j];
				int currentY = tempY.get(0)+neighborsY[j];
				if( currentX < model.sizeX && currentX >= 0 && currentY < model.sizeY && currentY >= 0) {
					double neighborDist = distanceToRoads.getValueAt(currentX, currentY);
					if(neighborDist > currentDistance + 1) { //if this cell is one step further away from what our current cell used to be
						tempX.add(currentX); //add it to the list of things to get modified
						tempY.add(currentY);
						distanceToRoads.putValueAt(currentX, currentY, currentDistance+1);
						closestRoadNode.putValueAt(currentX, currentY, currentRoadNode); //if this distance gets changed by adding road node x, then clearly node x is the closest one.
					}
				}
			}
			tempX.remove(0); //get rid of the one we just finished looking at
			tempY.remove(0);
		}
		
    }
    
    //evalDistance()
    public void evalDistance() {
    	//reset all distances
		for(int i = 0; i < model.sizeX; i++) {
			for(int j = 0; j < model.sizeY; j++) {
				distanceToRoads.putValueAt(i, j, -9999);		
			}
		}
		//start with a list of current roads
    	Vector<Double> a = distanceToRoads.findMinimum((int) model.sizeX/2, (int) model.sizeY/2, (int) model.sizeX/2, true, 1);
    	//Vector<Integer> tempX = minorRoadListX;
    	//Vector<Integer> tempY = minorRoadListY;
       	Vector<Integer> tempX = roadListX;
    	Vector<Integer> tempY = roadListY;
    	int distance = 0;
    	//work your way out from the roads
    	if(tempX.size() > 0)  {//if there are any minor roads yet...
    		int m = 0;
    		while ((Double) a.elementAt(0) < 0) {  //while there are still cells marked -9999
    			m++;
    			if(m > 10000) {
    				int poo = 1;
    				poo = 2;
    			}
    			Vector<Integer> newTempX = new Vector<Integer>();
    			Vector<Integer> newTempY = new Vector<Integer>();
    			for(int i = 0; i < tempX.size(); ++i){
    				Integer currentX = (Integer) tempX.get(i);
    				Integer currentY = (Integer) tempY.get(i);
    				//mark cells with current distance
    				distanceToRoads.putValueAt(currentX, currentY, distance);
    				
    				//mark which road node is closest;
    				//Vector<Double> closestNode = (Vector<Double>) closestRoadNode.findMinimum(currentX, currentY, 1, false, Discrete2DSpace.VON_NEUMANN);
    				//Vector<Double> vnNeighbors = (Vector<Double>) closestRoadNode.getVonNeumannNeighbors(currentX, currentY, false);//(currentX, currentY, 1, true, Discrete2DSpace.VON_NEUMANN);
    				//int closestNode = 999999999;
    				//for(int m = 0; m < vnNeighbors.size(); m++) {
    				//	closestNode = (int) Math.min( closestNode, vnNeighbors.get(m));
    				//}
    				//closestRoadNode.putValueAt(currentX, currentY, closestNode);
    			}
    			//add neighboring unmarked cells to list of cells to try next
    			for(int i = 0; i < tempX.size(); ++i){
    				Integer currentX = (Integer) tempX.get(i);
    				Integer currentY = (Integer) tempY.get(i);

    				if( currentX > 0) {
    					if(distanceToRoads.getValueAt(currentX-1, currentY) < -5000) { //if not added to the list
    						Integer newX = currentX-1;
    						Integer newY = currentY;
    						newTempX.add(newX);
    						newTempY.add(newY);
    						distanceToRoads.putValueAt(currentX-1, currentY, -4999); //we've added this one to the list, but haven't calculated it yet
    					}
    				}
    				if(currentX < model.sizeX-1) {
    					if(distanceToRoads.getValueAt(currentX+1, currentY) < -5000) {
    						Integer newX = currentX+1;
    						Integer newY = currentY;
    						newTempX.add(newX);
    						newTempY.add(newY);		
    						distanceToRoads.putValueAt(currentX+1, currentY, -4999);
    					}
    				}
    				if(currentY < model.sizeY-1) {
    					if(distanceToRoads.getValueAt(currentX, currentY+1) < -5000) {
    						Integer newX = currentX;
    						Integer newY = currentY+1;
    						newTempX.add(newX);
    						newTempY.add(newY);	
    						distanceToRoads.putValueAt(currentX, currentY+1, -4999);
    					}
    				}
    				if(currentY > 0) {
    					if(distanceToRoads.getValueAt(currentX, currentY-1) < -5000) {
    						Integer newX = currentX;
    						Integer newY = currentY-1;
    						newTempX.add(newX);
    						newTempY.add(newY);	
    						distanceToRoads.putValueAt(currentX, currentY-1, -4999);
    					}
    				}
    			}

    			//tempX.clear();
    			//tempY.clear();
    			tempX = (Vector<Integer>) newTempX.clone();
    			tempY = (Vector<Integer>) newTempY.clone();
    			distance++;
    			a = distanceToRoads.findMinimum((int) model.sizeX/2, (int) model.sizeY/2, (int) model.sizeX/2, true, 1);
    			newTempX.clear();
    			newTempY.clear();
    		}

    	}
    	tempX.clear();
    	tempY.clear();
    	a.clear();
    	
    	//collect statistics
    	double tempPercentPaved = 0;
    	double tempAveDistRoad = 0;
		for(int i = 0; i < model.sizeX; i++) {
			for(int j = 0; j < model.sizeY; j++) {
				tempAveDistRoad += distanceToRoads.getValueAt(i, j);
				if (roads.getValueAt(i, j) > 0) {
					tempPercentPaved += 1;
				}
			}
		}
    	percentPaved = tempPercentPaved/model.sizeX/model.sizeY;
    	aveDistRoad = tempAveDistRoad/model.sizeX/model.sizeY;
    }
    
    public void collectStatistics() {
    	//collect statistics
    	double tempPercentPaved = 0;
    	double tempAveDistRoad = 0;
		for(int i = 0; i < model.sizeX; i++) {
			for(int j = 0; j < model.sizeY; j++) {
				tempAveDistRoad += distanceToRoads.getValueAt(i, j);
				if (roads.getValueAt(i, j) > 0) {
					tempPercentPaved += 1;
				}
			}
		}
    	percentPaved = tempPercentPaved/model.sizeX/model.sizeY;
    	aveDistRoad = tempAveDistRoad/model.sizeX/model.sizeY;
    }
    
 

	////////////////////////////////////////////////////////////////////////////
	// setters and getters
	public void setID(int i) {	
		ID = i; 
	}
	public int getID() {
		return ID;
	}
	

	//Class Methods
	public static void setNextId ( int i ) { nextID = i; }

	public static void setWorld( GridWorld w ) {	world = w; }
	public static void setModel( Model m ) { model = m; }

}