/**
 * Tree.java 
 */

package CAFI;

import uchicago.src.sim.gui.*;
import uchicago.src.sim.space.Object2DGrid;
import java.util.Vector;

public class Tree {
	// class variables
	public  static int          	nextID = 0; // to give each an ID
	public  static Model	model;
	public  static GridWorld world;


	// instance variables  
	int 				ID;
	public Object2DGrid volume;
	public double		totalBiomass;
	public Object2DGrid tempVolume;

	// the Tree constructor
	public Tree (double maxVolume, int speciesNum) {
		ID = nextID++;

		volume = new Object2DGrid ( model.sizeX, model.sizeY);
		totalBiomass = 0;

		for(int i = 0; i < model.sizeX; i++) {
			for(int j = 0; j < model.sizeY; j++) {

				//just a simple pattern, for testing
				//double randVolume = maxVolume*i*j/model.sizeX/model.sizeY;
				double randVolume = 0;
				if(model.treeDistTestOn) {
					randVolume = maxVolume/2;
					/*if(speciesNum == 1) {
						if(i < (double) model.sizeX/2) {
							randVolume = maxVolume; //model.getUniformDoubleFromTo(1, maxVolume);
						}else {
							randVolume = 0;
						}
					}else if(speciesNum == 2) {
						if(i < (double) model.sizeX/2) {
							randVolume = maxVolume; // model.getUniformDoubleFromTo(1, maxVolume);
						}else {
							randVolume = maxVolume*2; //model.getUniformDoubleFromTo(1, maxVolume*2);
						}
					}else {
						randVolume = maxVolume; // model.getUniformDoubleFromTo(1, maxVolume);
					}*/
				} else {
					if(model.usePowerDist){
						randVolume = maxVolume/2*Math.pow(1-model.getUniformDoubleFromTo(0, 1), (-1/(model.powerAlpha-1)));  
					} else {
						randVolume = model.getUniformDoubleFromTo(1, maxVolume);
					}
				}

				volume.putValueAt(i,j, randVolume);
				totalBiomass += randVolume;
			}
		}

		if(model.usePowerDist) {
			double scalingFactor = (model.sizeX * model.sizeY * maxVolume/2) / (totalBiomass);
			for(int i = 0; i < model.sizeX; i++) {
				for(int j = 0; j < model.sizeY; j++) {
					double currentVolume = volume.getValueAt(i, j);
					volume.putValueAt(i, j, currentVolume * scalingFactor);
				}
			}
		}
		if(model.useKernelSmooth) {
			double halfRadius = (double) model.kernelRadius/2;

			//make a little grid where we can check if we have already counted a cell.  the center of this little
			//grid corresponds to the current point being generated
			Object2DGrid tempCheck = new Object2DGrid(model.kernelRadius*2+1,model.kernelRadius*2+1);
			Vector<Integer> tempX = new Vector<Integer> (0);
			Vector<Integer> tempY = new Vector<Integer> (0);
			Vector<Integer> tempSmallX = new Vector<Integer> (0);
			Vector<Integer> tempSmallY = new Vector<Integer> (0);

			//do the kerneling
			tempVolume = new Object2DGrid(model.sizeX, model.sizeY );
			for(int i = 0; i < model.sizeX; i++) {
				for(int j = 0; j < model.sizeY; j++) {
					double curveDistance = -halfRadius;
					for(int iTemp = 0; iTemp < tempCheck.getSizeX(); iTemp++) {
						for(int jTemp = 0; jTemp < tempCheck.getSizeY(); jTemp++) {
							tempCheck.putValueAt(iTemp, jTemp, 0);
						}
					}
					tempSmallX.add(model.kernelRadius);
					tempSmallY.add(model.kernelRadius);
					tempX.add(i);
					tempY.add(j);
					tempCheck.putValueAt(model.kernelRadius, model.kernelRadius, 1);
					double weightSum = 0;
					double fullSum = 0;
					for(int distance = 0; distance <= model.kernelRadius; distance++) {
						double weight = 1/(1+Math.exp(model.kernelWeightParam*curveDistance));
						Vector<Integer> newTempX = new Vector<Integer> (0);
						Vector<Integer> newTempY = new Vector<Integer> (0);
						Vector<Integer> newTempSmallX = new Vector<Integer> (0);
						Vector<Integer> newTempSmallY = new Vector<Integer> (0);
						for(int iList = 0; iList < tempX.size(); iList++) {
							int currentX = tempX.get(iList);
							int currentY = tempY.get(iList);
							int currentSmallX = tempSmallX.get(iList);
							int currentSmallY = tempSmallY.get(iList);
							weightSum += weight;
							fullSum += weight * volume.getValueAt(tempX.get(iList), tempY.get(iList));
							if(distance < model.kernelRadius) {
								//add in the search for the next layer out
								if(currentX > 0) {
									if(tempCheck.getValueAt(currentSmallX-1, currentSmallY) == 0) {

										newTempX.add(currentX-1);
										newTempY.add(currentY);
										newTempSmallX.add(currentSmallX-1);
										newTempSmallY.add(currentSmallY);
										tempCheck.putValueAt(currentSmallX-1, currentSmallY, 1);
									}
								}
								if(currentX < model.sizeX-1) {
									if(tempCheck.getValueAt(currentSmallX+1, currentSmallY) == 0) {
										newTempX.add(currentX+1);
										newTempY.add(currentY);
										newTempSmallX.add(currentSmallX+1);
										newTempSmallY.add(currentSmallY);
										tempCheck.putValueAt(currentSmallX+1, currentSmallY, 1);
									}
								}
								if(currentY > 0) {
									if(tempCheck.getValueAt(currentSmallX, currentSmallY-1) == 0) {
										newTempX.add(currentX);
										newTempY.add(currentY-1);
										newTempSmallX.add(currentSmallX);
										newTempSmallY.add(currentSmallY-1);
										tempCheck.putValueAt(currentSmallX, currentSmallY-1, 1);
									}
								}
								if(currentY < model.sizeY-1) {
									if(tempCheck.getValueAt(currentSmallX, currentSmallY+1) == 0) {
										newTempX.add(currentX);
										newTempY.add(currentY+1);
										newTempSmallX.add(currentSmallX);
										newTempSmallY.add(currentSmallY+1);
										tempCheck.putValueAt(currentSmallX, currentSmallY+1, 1);
									}
								}
							}
						}

						tempX = newTempX;
						tempY = newTempY;
						tempSmallX = newTempSmallX;
						tempSmallY = newTempSmallY;
						curveDistance += 1;
					}

					double weightedVolume = fullSum / weightSum;
					tempVolume.putValueAt(i, j, weightedVolume);

				}
			}

			volume = tempVolume;
			updateBiomass();
		}


	}

	//update biomass
	public void updateBiomass () {
		totalBiomass = 0;

		for(int i = 0; i < model.sizeX; i++) {
			for(int j = 0; j < model.sizeY; j++) {

				totalBiomass += volume.getValueAt(i, j);
			}
		}
	}

	////////////////////////////////////////////////////////////////////////////
	// setters and getters
	public void setID(int i) {	
		ID = i; 
	}
	public int getID() {
		return ID;
	}


	//Class Methods
	public static void setNextId ( int i ) { nextID = i; }

	public static void setWorld( GridWorld w ) {	world = w; }
	public static void setModel( Model m ) { model = m; }

}